import logging
import os
import warnings
from functools import wraps

import numpy as np

# Optimize imported here and below as the jax-optimized one is jax or passthrough, but this is required regardless
import scipy.optimize
from pymbar.utils import ensure_type, check_w_normalized, ParameterError

logger = logging.getLogger(__name__)

use_jit = False


# Temporary until we can make a proper setting to enable/disable by choice at runtime
def _setup_jax_acceleration():
    return os.environ.get("PYMBAR_DISABLE_JAX", "").lower() in ("true", "yes", "1")


# Setup if we should use jax or not
force_no_jax = _setup_jax_acceleration()

try:
    #### JAX related imports
    if force_no_jax:
        # Capture user-disabled JAX instead "JAX not found"
        raise ImportError("Jax disabled by force_no_jax in mbar_solvers.py")
    try:
        # This works with pre- and post- JAX 0.3.25
        # If you use try...except around a from jax.config import config then
        # config gets mangled and the x64_enabled flag doesn't exist.
        from jax import config

        if not config.x64_enabled:
            # Warn that we're going to be setting 64 bit jax
            logger.warning(
                "\n"
                "****** PyMBAR will use 64-bit JAX! *******\n"
                "* JAX is currently set to 32-bit bitsize *\n"
                "* which is its default.                  *\n"
                "*                                        *\n"
                "* PyMBAR requires 64-bit mode and WILL   *\n"
                "* enable JAX's 64-bit mode when called.  *\n"
                "*                                        *\n"
                "* This MAY cause problems with other     *\n"
                "* Uses of JAX in the same code.          *\n"
                "******************************************\n"
            )

        from jax.numpy import exp, sum, newaxis, diag, dot, s_
        from jax.numpy import pad as npad
        from jax.numpy.linalg import lstsq
        import jax.scipy.optimize as optimize_maybe_jax
        from jax.scipy.special import logsumexp

        from jax import jit as jit_or_passthrough

        use_jit = True
    except ImportError:
        # Catch no JAX and throw a warning
        logger.warning(
            "\n"
            "********* JAX NOT FOUND *********\n"
            " PyMBAR can run faster with JAX  \n"
            " But will work fine without it   \n"
            "Either install with pip or conda:\n"
            "      pip install pymbar[jax]     \n"
            "               OR                \n"
            "      conda install pymbar       \n"
            "*********************************"
        )
        raise  # Continue with the raised Import Error

except ImportError:
    # No JAX found, overlap imports
    # These imports MUST align exactly
    from numpy import exp, sum, newaxis, diag, dot, s_
    from numpy import pad as npad
    from numpy.linalg import lstsq
    import scipy.optimize as optimize_maybe_jax  # pylint: disable=reimported
    from scipy.special import logsumexp

    # No jit, so make a passthrough decorator
    def jit_or_passthrough(fn):
        return fn


# Note on "pylint: disable=invalid-unary-operand-type"
# Known issue with astroid<2.12 and numpy array returns, but 2.12 doesn't fix it due to returns being jax.
# Can be mostly ignored

if use_jit is False:
    logger.info("JAX was either not detected or disabled, using standard NumPy and SciPy")
else:
    logger.info("JAX detected. Using JAX acceleration.")

# Below are the recommended default protocols (ordered sequence of minimization algorithms / NLE solvers) for solving
# the MBAR equations.
# Note: we use tuples instead of lists to avoid accidental mutability.
JAX_SOLVER_PROTOCOL = (
    dict(method="BFGS", continuation=True),
    dict(method="adaptive", options=dict(min_sc_iter=0)),
)

DEFAULT_SOLVER_PROTOCOL = (
    dict(method="hybr", continuation=True),
    dict(method="adaptive", options=dict(min_sc_iter=0)),
)

ROBUST_SOLVER_PROTOCOL = (
    dict(method="adaptive", options=dict(maxiter=1000)),
    dict(method="L-BFGS-B", options=dict(maxiter=1000)),
)

BOOTSTRAP_SOLVER_PROTOCOL = (dict(method="adaptive", options=dict(min_sc_iter=0)),)

# Allows all of the gradient based methods, but not the non-gradient methods ["Nelder-Mead", "Powell", "COBYLA"]",
scipy_minimize_options = [
    "L-BFGS-B",
    "dogleg",
    "CG",
    "BFGS",
    "Newton-CG",
    "TNC",
    "trust-ncg",
    "trust-krylov",
    "trust-exact",
    "SLSQP",
]
scipy_nohess_options = [
    "L-BFGS-B",
    "BFGS",
    "CG",
    "TNC",
    "SLSQP",
]  # don't pass a hessian to these to avoid warnings to these.
scipy_root_options = ["hybr", "lm"]  # only use root options with the hessian included


def jit_or_pass_after_bitsize(jitable_fn):
    """
    Attempt to set JAX precision if present. This does nothing if JAX is not present

    Parameters
    ----------
    jitable_fn: function
        A function which can be jit'd
    """

    @wraps(
        jitable_fn
    )  # Helper to ensure the decorated function still registers for docs and inspection
    def staggered_jit(*args, **kwargs):
        # This will only trigger if JAX is set
        if use_jit and not config.x64_enabled:
            # Warn that JAX 64-bit will being turned on
            logger.warning(
                "\n"
                "******* JAX 64-bit mode is now on! *******\n"
                "*     JAX is now set to 64-bit mode!     *\n"
                "*   This MAY cause problems with other   *\n"
                "*      uses of JAX in the same code.     *\n"
                "******************************************\n"
            )
            config.update("jax_enable_x64", True)
        jited_fn = jit_or_passthrough(jitable_fn)
        return jited_fn(*args, **kwargs)

    return staggered_jit


def validate_inputs(u_kn, N_k, f_k):
    """Check types and return inputs for MBAR calculations.

    Parameters
    ----------
    u_kn or q_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies or unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    u_kn or q_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies or unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='float'
        The number of samples in each state.  Converted to float because this cast is required when log is calculated.
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state
    """
    n_states, n_samples = u_kn.shape

    u_kn = ensure_type(u_kn, "float", 2, "u_kn or Q_kn", shape=(n_states, n_samples))
    N_k = ensure_type(
        N_k, "float", 1, "N_k", shape=(n_states,), warn_on_cast=False
    )  # Autocast to float because will be eventually used in float calculations.
    f_k = ensure_type(f_k, "float", 1, "f_k", shape=(n_states,))

    return u_kn, N_k, f_k


def self_consistent_update(u_kn, N_k, f_k, states_with_samples=None):
    """Return an improved guess for the dimensionless free energies

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    f_k : np.ndarray, shape=(n_states), dtype='float'
        Updated estimate of f_k

    Notes
    -----
    Equation C3 in MBAR JCP paper.
    """

    return jax_self_consistent_update(u_kn, N_k, f_k, states_with_samples=states_with_samples)


@jit_or_pass_after_bitsize
def _jit_self_consistent_update(u_kn, N_k, f_k):
    """JAX version of self_consistent update.  For parameters, see self_consistent_update.
    N_k must be float (should be cast at a higher level)

    """
    # Asteroid
    log_denominator_n = logsumexp(f_k - u_kn.T, b=N_k, axis=1)
    # All states can contribute to the numerator term. Check transpose
    return -1.0 * logsumexp(
        -log_denominator_n - u_kn, axis=1
    )  # pylint: disable=invalid-unary-operand-type


def jax_self_consistent_update(u_kn, N_k, f_k, states_with_samples=None):
    """JAX version of self_consistent update.  For parameters, see self_consistent_update.
    N_k must be float (should be cast at a higher level)

    """
    # Only the states with samples can contribute to the denominator term.
    # Precondition before feeding the op to the JIT'd function
    # In theory, this can be computed with jax.lax.cond, but trying to reuse code for non-jax paths
    states_with_samples = s_[:] if states_with_samples is None else states_with_samples
    # Feed to the JIT'd function. Can't pass slice types, so slice here
    return _jit_self_consistent_update(
        u_kn[states_with_samples], N_k[states_with_samples], f_k[states_with_samples]
    )


def mbar_gradient(u_kn, N_k, f_k):
    """Gradient of MBAR objective function.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    grad : np.ndarray, dtype=float, shape=(n_states)
        Gradient of mbar_objective

    Notes
    -----
    This is equation C6 in the JCP MBAR paper.
    """
    return jax_mbar_gradient(u_kn, N_k, f_k)


@jit_or_pass_after_bitsize
def jax_mbar_gradient(u_kn, N_k, f_k):
    """JAX version of MBAR gradient function. See documentation of mbar_gradient.
    N_k must be float (should be cast at a higher level)
    """

    log_denominator_n = logsumexp(f_k - u_kn.T, b=N_k, axis=1)
    log_numerator_k = logsumexp(-log_denominator_n - u_kn, axis=1)
    return -1 * N_k * (1.0 - exp(f_k + log_numerator_k))


def mbar_objective(u_kn, N_k, f_k):
    """Calculates objective function for MBAR.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state


    Returns
    -------
    obj : float
        Objective function

    Notes
    -----
    This objective function is essentially a doubly-summed partition function and is
    quite sensitive to precision loss from both overflow and underflow. For optimal
    results, u_kn can be preconditioned by subtracting out a `n` dependent
    vector.

    More optimal precision, the objective function uses math.fsum for the
    outermost sum and logsumexp for the inner sum.
    """

    return jax_mbar_objective(u_kn, N_k, f_k)


@jit_or_pass_after_bitsize
def jax_mbar_objective(u_kn, N_k, f_k):
    """JAX version of mbar_objective.
    For parameters, mbar_objective_and_Gradient
    N_k must be float (should be cast at a higher level)

    """

    log_denominator_n = logsumexp(f_k - u_kn.T, b=N_k, axis=1)
    obj = sum(log_denominator_n) - dot(N_k, f_k)

    return obj


@jit_or_pass_after_bitsize
def jax_mbar_objective_and_gradient(u_kn, N_k, f_k):
    """JAX version of mbar_objective_and_gradient.
    For parameters, mbar_objective_and_Gradient
    N_k must be float (should be cast at a higher level)

    """

    log_denominator_n = logsumexp(f_k - u_kn.T, b=N_k, axis=1)
    log_numerator_k = logsumexp(-log_denominator_n - u_kn, axis=1)
    grad = -1 * N_k * (1.0 - exp(f_k + log_numerator_k))

    obj = sum(log_denominator_n) - dot(N_k, f_k)

    return obj, grad


def mbar_objective_and_gradient(u_kn, N_k, f_k):
    """Calculates both objective function and gradient for MBAR.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state


    Returns
    -------
    obj : float
        Objective function
    grad : np.ndarray, dtype=float, shape=(n_states)
        Gradient of objective function

    Notes
    -----
    This objective function is essentially a doubly-summed partition function and is
    quite sensitive to precision loss from both overflow and underflow. For optimal
    results, u_kn can be preconditioned by subtracting out a `n` dependent
    vector.

    More optimal precision, the objective function uses math.fsum for the
    outermost sum and logsumexp for the inner sum.

    The gradient is equation C6 in the JCP MBAR paper; the objective
    function is its integral.
    """

    return jax_mbar_objective_and_gradient(u_kn, N_k, f_k)


@jit_or_pass_after_bitsize
def jax_mbar_hessian(u_kn, N_k, f_k):
    """JAX version of mbar_hessian.
    For parameters, see mbar_hessian
    N_k must be float (should be cast at a higher level)

    """

    log_denominator_n = logsumexp(f_k - u_kn.T, b=N_k, axis=1)
    logW = f_k - u_kn.T - log_denominator_n[:, newaxis]
    W = exp(logW)

    H = dot(W.T, W)
    H *= N_k
    H *= N_k[:, newaxis]
    H -= diag(W.sum(0) * N_k)
    return -1.0 * H


def mbar_hessian(u_kn, N_k, f_k):
    """Hessian of MBAR objective function.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    H : np.ndarray, dtype=float, shape=(n_states, n_states)
        Hessian of mbar objective function.

    Notes
    -----
    Equation (C9) in JCP MBAR paper.
    """

    return jax_mbar_hessian(u_kn, N_k, f_k)


@jit_or_pass_after_bitsize
def jax_mbar_log_W_nk(u_kn, N_k, f_k):
    """JAX version of mbar_log_W_nk.
    For parameters, see mbar_log_W_nk
    N_k must be float (should be cast at a higher level)

    """

    log_denominator_n = logsumexp(f_k - u_kn.T, b=N_k, axis=1)
    logW = f_k - u_kn.T - log_denominator_n[:, newaxis]
    return logW


def mbar_log_W_nk(u_kn, N_k, f_k):
    """Calculate the log weight matrix.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    logW_nk : np.ndarray, dtype='float', shape=(n_samples, n_states)
        The normalized log weights.

    Notes
    -----
    Equation (9) in JCP MBAR paper.
    """
    return jax_mbar_log_W_nk(u_kn, N_k, f_k)


@jit_or_pass_after_bitsize
def jax_mbar_W_nk(u_kn, N_k, f_k):
    """JAX version of mbar_W_nk.
    For parameters, see mbar_W_nk
    N_k must be float (should be cast at a higher level)

    """
    return exp(jax_mbar_log_W_nk(u_kn, N_k, f_k))


def mbar_W_nk(u_kn, N_k, f_k):
    """Calculate the weight matrix.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    W_nk : np.ndarray, dtype='float', shape=(n_samples, n_states)
        The normalized weights.

    Notes
    -----
    Equation (9) in JCP MBAR paper.
    """
    return jax_mbar_W_nk(u_kn, N_k, f_k)


def adaptive(u_kn, N_k, f_k, tol=1.0e-8, options=None):
    """
    Determine dimensionless free energies by a combination of Newton-Raphson iteration and self-consistent iteration.
    Picks whichever method gives the lowest gradient.
    Is slower than NR since it calculates the log norms twice each iteration.

    OPTIONAL ARGUMENTS
    tol (float between 0 and 1) - relative tolerance for convergence (default 1.0e-12)

    options : dictionary of options
        gamma (float between 0 and 1) - incrementor for NR iterations (default 1.0).  Usually not changed now, since adaptively switch.
        maxiter (int) - maximum number of Newton-Raphson iterations (default 10000: either NR converges or doesn't, pretty quickly)
        verbose (boolean) - verbosity level for debug output

    NOTES

    This method determines the dimensionless free energies by
    minimizing a convex function whose solution is the desired
    estimator.  The original idea came from the construction of a
    likelihood function that independently reproduced the work of
    Geyer (see [1] and Section 6 of [2]).  This can alternatively be
    formulated as a root-finding algorithm for the Z-estimator.  More
    details of this procedure will follow in a subsequent paper.  Only
    those states with nonzero counts are include in the estimation
    procedure.

    REFERENCES
    See Appendix C.2 of [1].

    """
    # put the defaults here in case we get passed an 'options' dictionary that is only partial
    options.setdefault("verbose", False)
    options.setdefault("maxiter", 10000)
    options.setdefault("print_warning", False)
    options.setdefault("gamma", 1.0)
    options.setdefault("min_sc_iter", 2)  # set a minimum number of self-consistent iterations

    gamma = options["gamma"]

    doneIterating = False
    if options["verbose"] == True:
        logger.info(
            "Determining dimensionless free energies by Newton-Raphson / self-consistent iteration."
        )

    if tol < 4.0 * np.finfo(float).eps:
        logger.info("Tolerance may be too close to machine precision to converge.")

    success = False  # fail unless solution is found.
    # keep track of Newton-Raphson and self-consistent iterations
    nr_iter = 0
    sci_iter = 0

    f_sci = np.zeros(len(f_k), dtype=np.float64)
    f_nr = np.zeros(len(f_k), dtype=np.float64)

    # Perform Newton-Raphson iterations (with sci computed on the way)

    # usually calculated at the end of the loop and saved, but we need
    # to calculate the first time.
    g = mbar_gradient(u_kn, N_k, f_k)  # Objective function gradient.

    maxiter = options["maxiter"]
    min_sc_iter = options["min_sc_iter"]
    warn = "Did not converge."
    for iteration in range(0, maxiter):
        if use_jit:
            (f_sci, g_sci, gnorm_sci, f_nr, g_nr, gnorm_nr) = jax_core_adaptive(
                u_kn, N_k, f_k, options["gamma"]
            )
        else:
            H = mbar_hessian(u_kn, N_k, f_k)  # Objective function hessian
            Hinvg = np.linalg.lstsq(H, g, rcond=-1)[0]
            Hinvg -= Hinvg[0]
            f_nr = f_k - gamma * Hinvg

            # self-consistent iteration gradient norm and saved log sums.
            f_sci = self_consistent_update(u_kn, N_k, f_k)
            f_sci = f_sci - f_sci[0]  # zero out the minimum
            g_sci = mbar_gradient(u_kn, N_k, f_sci)
            gnorm_sci = dot(g_sci, g_sci)

            # newton raphson gradient norm and saved log sums.
            g_nr = mbar_gradient(u_kn, N_k, f_nr)
            gnorm_nr = dot(g_nr, g_nr)

        # we could save the gradient, for the next round, but it's not too expensive to
        # compute since we are doing the Hessian anyway.

        if options["verbose"]:
            logger.info(
                "self consistent iteration gradient norm is %10.5g, Newton-Raphson gradient norm is %10.5g"
                % (np.sqrt(gnorm_sci), np.sqrt(gnorm_nr))
            )
        # decide which directon to go depending on size of gradient norm
        f_old = f_k

        if gnorm_sci < gnorm_nr or sci_iter < min_sc_iter:
            f_k = f_sci
            g = g_sci
            sci_iter += 1
            if options["verbose"]:
                if sci_iter < min_sc_iter:
                    logger.info(
                        f"Choosing self-consistent iteration on iteration {iteration:d} because min_sci_iter={min_sc_iter:d}"
                    )
                else:
                    logger.info(
                        f"Choosing self-consistent iteration for lower gradient on iteration {iteration:d}"
                    )
        else:
            f_k = f_nr
            g = g_nr
            nr_iter += 1
            if options["verbose"]:
                logger.info(f"Newton-Raphson used on iteration {iteration:}")

        div = np.abs(f_k[1:])  # what we will divide by to get relative difference
        zeroed = np.abs(f_k[1:]) < np.min(
            [10**-8, tol]
        )  # check which values are near enough to zero, hard coded max for now.
        div[zeroed] = 1.0  # for these values, use absolute values.
        max_delta = np.max(np.abs(f_k[1:] - f_old[1:]) / div)
        max_diff = np.max(np.abs(f_sci[1:] - f_nr[1:]) / div)
        # add this just to make sure they are not too different.
        # if we start with bad states, the f_k - f_k_old might be far off.
        if np.isnan(max_delta) or ((max_delta < tol) and max_diff < np.sqrt(tol)):
            doneIterating = True
            success = True
            warn = "Convergence achieved by change in f with respect to previous guess."
            break

    if doneIterating:
        if options["verbose"]:
            logger.info(f"Converged to tolerance of {max_delta:e} in {iteration+1:d} iterations.")
            logger.info(
                f"Of {iteration+1:d} iterations, {nr_iter:d} were Newton-Raphson iterations and {sci_iter:d} were self-consistent iterations"
            )
            if np.all(f_k == 0.0):
                logger.info("WARNING: All f_k appear to be zero.")
    else:
        logger.warning("WARNING: Did not converge to within specified tolerance.")

        if maxiter <= 0:
            logger.warning(
                f"No iterations ran be cause maximum_iterations was <= 0 ({maxiter:s})!"
            )
        else:
            logger.warning(
                f"max_delta = {max_delta:e}, tol = {tol:e}, maximum_iterations = {maxiter:d}, iterations completed = {iteration:d}"
            )

    results = dict()
    results["success"] = success
    results["message"] = warn
    results["x"] = f_k

    return results


@jit_or_pass_after_bitsize
def jax_core_adaptive(u_kn, N_k, f_k, gamma):
    """JAX version of adaptive inner loop.
    N_k must be float (should be cast at a higher level)

    """

    # Perform Newton-Raphson iterations (with sci computed on the way)
    g = mbar_gradient(u_kn, N_k, f_k)  # Objective function gradient
    H = mbar_hessian(u_kn, N_k, f_k)  # Objective function hessian
    Hinvg = lstsq(H, g, rcond=-1)[0]
    Hinvg -= Hinvg[0]
    f_nr = f_k - gamma * Hinvg

    # self-consistent iteration gradient norm and saved log sums.
    f_sci = self_consistent_update(u_kn, N_k, f_k)
    f_sci = f_sci - f_sci[0]  # zero out the minimum
    g_sci = mbar_gradient(u_kn, N_k, f_sci)
    gnorm_sci = dot(g_sci, g_sci)

    # newton raphson gradient norm and saved log sums.
    g_nr = mbar_gradient(u_kn, N_k, f_nr)
    gnorm_nr = dot(g_nr, g_nr)

    return f_sci, g_sci, gnorm_sci, f_nr, g_nr, gnorm_nr


@jit_or_pass_after_bitsize
def jax_precondition_u_kn(u_kn, N_k, f_k):
    """JAX version of precondition_u_kn
    for parameters, see precondition_u_kn
    N_k must be float (should be cast at a higher level)

    """

    u_kn = u_kn - u_kn.min(0)
    u_kn += (logsumexp(f_k - u_kn.T, b=N_k, axis=1)) - dot(N_k, f_k) / N_k.sum()
    return u_kn


def precondition_u_kn(u_kn, N_k, f_k):
    """Subtract a sample-dependent constant from u_kn to improve precision

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state

    Returns
    -------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities

    Notes
    -----
    Returns u_kn - x_n, where x_n is based on the current estimate of f_k.
    Upon subtraction of x_n, the MBAR objective function changes by an
    additive constant, but its derivatives remain unchanged.  We choose
    x_n such that the current objective function value is zero, which
    should give maximum precision in the objective function.
    """
    return jax_precondition_u_kn(u_kn, N_k, f_k)


def solve_mbar_once(
    u_kn_nonzero,
    N_k_nonzero,
    f_k_nonzero,
    method="adaptive",
    tol=1e-12,
    continuation=None,
    options=None,
):
    """Solve MBAR self-consistent equations using some form of equation solver.

    Parameters
    ----------
    u_kn_nonzero : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
        for the nonempty states
    N_k_nonzero : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state for the nonempty states
    f_k_nonzero : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies for the nonempty states
    method : str, optional, default="hybr"
        The optimization routine to use.  This can be any of the methods
        available via scipy.optimize.minimize() or scipy.optimize.root().
    tol : float, optional, default=1E-14
        The convergance tolerance for minimize() or root()
    verbose: bool
        Whether to print information about the solution method.
    options: dict, optional, default=None
        Optional dictionary of algorithm-specific parameters.  See
        scipy.optimize.root or scipy.optimize.minimize for details.

    Returns
    -------
    f_k : np.ndarray
        The converged reduced free energies.
    results : dict
        Dictionary containing entire results of optimization routine, may
        be useful when debugging convergence.

    Notes
    -----
    This function requires that N_k_nonzero > 0--that is, you should have
    already dropped all the states for which you have no samples.
    Internally, this function works in a reduced coordinate system defined
    by subtracting off the first component of f_k and fixing that component
    to be zero.

    For fast but precise convergence, we recommend calling this function
    multiple times to polish the result.  `solve_mbar()` facilitates this.
    """

    # we only validate at the outside of the call
    u_kn_nonzero, N_k_nonzeo, f_k_nonzero = validate_inputs(u_kn_nonzero, N_k_nonzero, f_k_nonzero)
    f_k_nonzero = f_k_nonzero - f_k_nonzero[0]  # Work with reduced dimensions with f_k[0] := 0
    N_k_nonzero = 1.0 * N_k_nonzero  # convert to float for acceleration.
    u_kn_nonzero = precondition_u_kn(u_kn_nonzero, N_k_nonzero, f_k_nonzero)

    pad = lambda x: np.pad(
        x, (1, 0), mode="constant"
    )  # Helper function inserts zero before first element
    unpad_second_arg = lambda obj, grad: (
        obj,
        grad[1:],
    )  # Helper function drops first element of gradient

    # Create objective functions / nonlinear equations to send to scipy.optimize, fixing f_0 = 0
    grad = lambda x: mbar_gradient(u_kn_nonzero, N_k_nonzero, pad(x))[
        1:
    ]  # Objective function gradient

    grad_and_obj = lambda x: unpad_second_arg(
        *mbar_objective_and_gradient(u_kn_nonzero, N_k_nonzero, pad(x))
    )  # Objective function gradient and objective function

    de_jax_grad_and_obj = lambda x: (
        *map(np.array, grad_and_obj(x)),  # (...,) Casts to tuple instead of <map> object
    )  # Force any jax-based array output to normal numpy for scipy.optimize.minimize. np.asarray does not work.

    hess = lambda x: mbar_hessian(u_kn_nonzero, N_k_nonzero, pad(x))[1:][
        :, 1:
    ]  # Hessian of objective function
    with warnings.catch_warnings(record=True) as w:
        if use_jit and method == "BFGS":
            fpad = lambda x: npad(x, (1, 0))
            obj = lambda x: mbar_objective(u_kn_nonzero, N_k_nonzero, fpad(x))
            # objective function to be minimized (for derivative free methods, mostly jit)
            jax_results = optimize_maybe_jax.minimize(
                obj,
                f_k_nonzero[1:],
                method=method,
                tol=tol,
                options=dict(maxiter=options["maxiter"]),
            )
            results = dict()  # there should be a way to copy this.
            results["x"] = jax_results[0]
            f_k_nonzero = pad(results["x"])
            results["success"] = jax_results[1]
        elif method in scipy_minimize_options:
            if method in scipy_nohess_options:
                hess = None  # To suppress warning from passing a hessian function.
            results = scipy.optimize.minimize(
                de_jax_grad_and_obj,
                f_k_nonzero[1:],
                jac=True,
                hess=hess,
                method=method,
                tol=tol,
                options=options,
            )
            f_k_nonzero = pad(results["x"])
        elif method == "adaptive":
            results = adaptive(u_kn_nonzero, N_k_nonzero, f_k_nonzero, tol=tol, options=options)
            f_k_nonzero = results["x"]
        elif method in scipy_root_options:
            # find the root in the gradient.
            results = scipy.optimize.root(
                grad, f_k_nonzero[1:], jac=hess, method=method, tol=tol, options=options
            )
            f_k_nonzero = pad(results["x"])
        else:
            raise ParameterError(f"Method {method} for solution of free energies not recognized")

    # If there were runtime warnings, show the messages
    if len(w) > 0:
        can_ignore = True
        for warn_msg in w:
            if "Unknown solver options" in str(warn_msg.message):
                continue
            warnings.showwarning(
                warn_msg.message,
                warn_msg.category,
                warn_msg.filename,
                warn_msg.lineno,
                warn_msg.file,
                "",
            )
            can_ignore = False  # If any warning is not just unknown options, can not skip check
        if not can_ignore:
            # Ensure MBAR solved correctly
            w_nk_check = mbar_W_nk(u_kn_nonzero, N_k_nonzero, f_k_nonzero)
            check_w_normalized(w_nk_check, N_k_nonzero)
            logger.warning(
                "MBAR weights converged within tolerance, despite the SciPy Warnings. Please validate your results."
            )

    return f_k_nonzero, results


def solve_mbar(u_kn_nonzero, N_k_nonzero, f_k_nonzero, solver_protocol=None):
    """Solve MBAR self-consistent equations using some sequence of equation solvers.

    Parameters
    ----------
    u_kn_nonzero : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
        for the nonempty states
    N_k_nonzero : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state for the nonempty states
    f_k_nonzero : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies for the nonempty states
    solver_protocol : tuple(dict()), optional, default=None
        Optional list of dictionaries of steps in solver protocol.
        If None, a default protocol will be used.

    Returns
    -------
    f_k : np.ndarray
        The converged reduced free energies.
    all_results : list(dict())
        List of results from each step of solver_protocol.  Each element in
        list contains the results dictionary from solve_mbar_once()
        for the corresponding step.

    Notes
    -----
    This function requires that N_k_nonzero > 0--that is, you should have
    already dropped all the states for which you have no samples.
    Internally, this function works in a reduced coordinate system defined
    by subtracting off the first component of f_k and fixing that component
    to be zero.

    This function calls `solve_mbar_once()` multiple times to achieve
    converged results.  Generally, a single call to solve_mbar_once()
    will not give fully converged answers because of limited numerical precision.
    Each call to `solve_mbar_once()` re-conditions the nonlinear
    equations using the current guess.
    """

    if solver_protocol is None:
        solver_protocol = DEFAULT_SOLVER_PROTOCOL

    all_fks = []
    all_gnorms = []
    all_results = []

    for solver in solver_protocol:
        f_k_nonzero_result, results = solve_mbar_once(
            u_kn_nonzero, N_k_nonzero, f_k_nonzero, **solver
        )
        all_fks.append(f_k_nonzero_result)
        all_gnorms.append(
            np.linalg.norm(mbar_gradient(u_kn_nonzero, N_k_nonzero, f_k_nonzero_result))
        )
        all_results.append(results)

        if results["success"]:
            success = True
            best_gnorm = all_gnorms[-1]
            logger.info(f"Reached a solution to within tolerance with {solver['method']}")
            break
        else:
            logger.warning(
                f"Failed to reach a solution to within tolerance with {solver['method']}: trying next method"
            )
        logger.info(f"Ending gnorm of method {solver['method']} = {all_gnorms[-1]:e}")
        if solver["continuation"]:
            f_k_nonzero = f_k_nonzero_result
            logger.info("Will continue with results from previous method")

    if results["success"]:
        logger.info("Solution found within tolerance!")
    else:
        i_best_gnorm = np.argmin(all_gnorms)
        logger.warning("No solution found to within tolerance.")
        best_method = solver_protocol[i_best_gnorm]["method"]
        best_gnorm = all_gnorms[i_best_gnorm]
        logger.warning(
            f"The solution with the smallest gradient {best_gnorm:e} norm is {best_method}"
        )
        f_k_nonzero_result = all_fks[i_best_gnorm]
        logger.warning(
            "Please exercise caution with this solution and consider alternative methods or a different tolerance."
        )

    logger.info(f"Final gradient norm: {best_gnorm:.3g}")

    return f_k_nonzero_result, all_results


def solve_mbar_for_all_states(u_kn, N_k, f_k, states_with_samples, solver_protocol):
    """Solve for free energies of states with samples, then calculate for
    empty states.

    Parameters
    ----------
    u_kn : np.ndarray, shape=(n_states, n_samples), dtype='float'
        The reduced potential energies, i.e. -log unnormalized probabilities
    N_k : np.ndarray, shape=(n_states), dtype='int'
        The number of samples in each state
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The reduced free energies of each state
    solver_protocol : tuple(dict()), optional, default=None
        Sequence of dictionaries of steps in solver protocol for final
        stage of refinement.

    Returns
    -------
    f_k : np.ndarray, shape=(n_states), dtype='float'
        The free energies of states
    """

    if len(states_with_samples) == 1:
        f_k_nonzero = np.array([0.0])
    else:
        f_k_nonzero, all_results = solve_mbar(
            u_kn[states_with_samples],
            N_k[states_with_samples],
            f_k[states_with_samples],
            solver_protocol=solver_protocol,
        )

    f_k[states_with_samples] = np.array(f_k_nonzero)

    # Update all free energies because those from states with zero samples are not correctly computed by solvers.
    f_k = self_consistent_update(u_kn, N_k, f_k)
    # This is necessary because state 0 might have had zero samples,
    # but we still want that state to be the reference with free energy 0.
    f_k -= f_k[0]

    return f_k



import numpy as np
import pytest
import pymbar
from pymbar.utils_for_testing import (
    assert_almost_equal,
    oscillators,
    exponentials,
)


@pytest.mark.parametrize(
    "statesa, statesb, test_system",
    [(100, 100, oscillators), (200, 50, oscillators), (200, 50, exponentials)],
)
def _test(statesa, statesb, test_system):
    name, U, N_k, s_n = test_system(statesa, statesb)
    print(name)
    mbar = pymbar.MBAR(U, N_k)
    results1 = mbar.compute_free_energy_differences(uncertainty_method="svd")
    results2 = mbar.compute_free_energy_differences(uncertainty_method="svd-ew")
    fij1 = results1["Delta_f"]
    dfij1 = results1["dDelta_f"]
    fij2 = results2["Delta_f"]
    dfij2 = results2["dDelta_f"]

    assert_almost_equal(
        pymbar.mbar_solvers.mbar_gradient(U, N_k, mbar.f_k), np.zeros(N_k.shape), decimal=8
    )
    assert_almost_equal(np.exp(mbar.Log_W_nk).sum(0), np.ones(len(N_k)), decimal=10)
    assert_almost_equal(np.exp(mbar.Log_W_nk).dot(N_k), np.ones(U.shape[1]), decimal=10)
    assert_almost_equal(
        pymbar.mbar_solvers.self_consistent_update(U, N_k, mbar.f_k), mbar.f_k, decimal=10
    )

    assert_almost_equal(fij1, fij2, decimal=8)
    assert_almost_equal(dfij1, dfij2, decimal=8)



from pymbar import timeseries
from pymbar import testsystems
import numpy as np
from scipy import stats
import pytest

from pymbar.utils_for_testing import assert_almost_equal

try:
    import statsmodels.api as sm  # pylint: disable=unused-import

    HAVE_STATSMODELS = True
except ImportError as err:
    HAVE_STATSMODELS = False

has_statmodels = pytest.mark.skipif(
    not HAVE_STATSMODELS, reason="Skipping FFT based tests because statsmodels not installed."
)


@pytest.fixture(scope="module")
def data(N=10000, K=10):
    var = np.ones(N)

    for replica in range(2, K + 1):
        var = np.concatenate((var, np.ones(N)))

    X = np.random.normal(np.zeros(K * N), var).reshape((K, N)) / 10.0
    Y = np.random.normal(np.zeros(K * N), var).reshape((K, N))

    energy = 10 * (X**2) / 2.0 + (Y**2) / 2.0

    return X, Y, energy


def test_statistical_inefficiency_single(data):
    X, Y, energy = data
    timeseries.statistical_inefficiency(X[0])
    timeseries.statistical_inefficiency(X[0], X[0])
    timeseries.statistical_inefficiency(X[0] ** 2)
    timeseries.statistical_inefficiency(X[0] ** 2, X[0] ** 2)
    timeseries.statistical_inefficiency(energy[0])
    timeseries.statistical_inefficiency(energy[0], energy[0])

    timeseries.statistical_inefficiency(X[0], X[0] ** 2)

    # TODO: Add some checks to test statistical inefficinecies are within normal range


def test_statistical_inefficiency_multiple(data):
    X, Y, energy = data
    timeseries.statistical_inefficiency_multiple(X)
    timeseries.statistical_inefficiency_multiple(X**2)
    timeseries.statistical_inefficiency_multiple(X[0, :] ** 2)
    timeseries.statistical_inefficiency_multiple(X[0:2, :] ** 2)
    timeseries.statistical_inefficiency_multiple(energy)
    # TODO: Add some checks to test statistical inefficinecies are within normal range


@has_statmodels
def test_statistical_inefficiency_fft(data):
    X, Y, energy = data
    timeseries.statistical_inefficiency_fft(X[0])
    timeseries.statistical_inefficiency_fft(X[0] ** 2)
    timeseries.statistical_inefficiency_fft(energy[0])

    g0 = timeseries.statistical_inefficiency_fft(X[0])
    g1 = timeseries.statistical_inefficiency(X[0])
    g2 = timeseries.statistical_inefficiency(X[0], X[0])
    g3 = timeseries.statistical_inefficiency(X[0], fft=True)
    assert_almost_equal(g0, g1, decimal=6)
    assert_almost_equal(g0, g2, decimal=6)
    assert_almost_equal(g0, g3, decimal=6)


@has_statmodels
def test_statistical_inefficiency_fft_gaussian():
    # Run multiple times to get things with and without negative "spikes" at C(1)
    for i in range(5):
        x = np.random.normal(size=100000)
        g0 = timeseries.statistical_inefficiency(x, fast=False)
        g1 = timeseries.statistical_inefficiency(x, x, fast=False)
        g2 = timeseries.statistical_inefficiency_fft(x)
        g3 = timeseries.statistical_inefficiency(x, fft=True)
        assert_almost_equal(g0, g1, decimal=5)
        assert_almost_equal(g0, g2, decimal=5)
        assert_almost_equal(g0, g3, decimal=5)

        assert_almost_equal(np.log(g0), np.log(1.0), decimal=1)

    for i in range(5):
        x = np.random.normal(size=100000)
        x = np.repeat(
            x, 3
        )  # e.g. Construct correlated gaussian e.g. [a, b, c] -> [a, a, a, b, b, b, c, c, c]
        g0 = timeseries.statistical_inefficiency(x, fast=False)
        g1 = timeseries.statistical_inefficiency(x, x, fast=False)
        g2 = timeseries.statistical_inefficiency_fft(x)
        g3 = timeseries.statistical_inefficiency(x, fft=True)
        assert_almost_equal(g0, g1, decimal=5)
        assert_almost_equal(g0, g2, decimal=5)
        assert_almost_equal(g0, g3, decimal=5)

        assert_almost_equal(np.log(g0), np.log(3.0), decimal=1)


def test_detectEquil():
    x = np.random.normal(size=10000)
    t, g, Neff_max = timeseries.detect_equilibration(x)


@has_statmodels
def test_detectEquil_binary():
    x = np.random.normal(size=10000)
    t, g, Neff_max = timeseries.detect_equilibration_binary_search(x)


@has_statmodels
def test_compare_detectEquil(show_hist=False):
    """
    compare detect_equilibration implementations (with and without binary search + fft)
    """
    t_res = []
    N = 100
    for _ in range(100):
        A_t = testsystems.correlated_timeseries_example(N=N, tau=5.0) + 2.0
        B_t = testsystems.correlated_timeseries_example(N=N, tau=5.0) + 1.0
        C_t = testsystems.correlated_timeseries_example(N=N * 2, tau=5.0)
        D_t = np.concatenate([A_t, B_t, C_t])
        bs_de = timeseries.detect_equilibration_binary_search(D_t, bs_nodes=10)
        std_de = timeseries.detect_equilibration(D_t, fast=False, nskip=1)
        t_res.append(bs_de[0] - std_de[0])
    try:
        # scipy<1.9
        t_res_mode = float(stats.mode(t_res)[0][0])
    except IndexError:
        # scipy>=1.9
        t_res_mode = float(stats.mode(t_res, keepdims=True)[0][0])
    assert_almost_equal(t_res_mode, 0.0, decimal=1)
    if show_hist:
        import matplotlib.pyplot as plt

        plt.hist(t_res)
        plt.show()


def test_detectEquil_constant_trailing():
    """
    This explicitly tests issue #122, see https://github.com/choderalab/pymbar/issues/122

    We only check that the code doesn't give an exception.  The exact value of Neff can either be
    ~50 if we try to include part of the equilibration samples, or it can be Neff=1 if we find that the
    whole first half is discarded.
    """
    x = np.random.normal(size=100) * 0.01
    x[50:] = 3.0
    # The input data is some MCMC chain where the trailing end of the chain is a constant sequence.
    t, g, Neff_max = timeseries.detect_equilibration(x)


def test_correlationFunctionMultiple():
    """
    tests the truncate and norm feature
    """
    A_t = [testsystems.correlated_timeseries_example(N=10000, tau=10.0) for _ in range(10)]
    corr_norm = timeseries.normalized_fluctuation_correlation_function_multiple(A_kn=A_t)
    corr = timeseries.normalized_fluctuation_correlation_function_multiple(A_kn=A_t, norm=False)
    corr_norm_trun = timeseries.normalized_fluctuation_correlation_function_multiple(
        A_kn=A_t, truncate=True
    )
    corr_trun = timeseries.normalized_fluctuation_correlation_function_multiple(
        A_kn=A_t, norm=False, truncate=True
    )
    assert corr_norm_trun[-1] >= 0
    assert corr_trun[-1] >= 0
    assert corr_norm[0] == 1.0
    assert corr_norm_trun[0] == 1.0
    assert len(corr_trun) == len(corr_norm_trun)



"""
Test MBAR by performing statistical tests on a set of of 1D harmonic oscillators, for which
the true free energy differences can be computed analytically.

A number of replications of an experiment in which i.i.d. samples are drawn from a set of
K harmonic oscillators are produced.  For each replicate, we estimate the dimensionless free
energy differences and mean-square displacements (an observable), as well as their uncertainties.

For a 1D harmonic oscillator, the potential is given by
  V(x;K) = (K/2) * (x-x_0)**2
where K denotes the spring constant.

The equilibrium distribution is given analytically by
  p(x;beta,K) = sqrt[(beta K) / (2 pi)] exp[-beta K (x-x_0)**2 / 2]
The dimensionless free energy is therefore
  f(beta,K) = - (1/2) * ln[ (2 pi) / (beta K) ]

"""

# =============================================================================================
# IMPORTS
# =============================================================================================

import sys
import numpy as np
from pymbar import testsystems, exp, exp_gauss, bar, MBAR, FES
from pymbar.utils import ParameterError

import logging
import sys

logging.basicConfig(stream=sys.stdout, level=logging.INFO)

# =============================================================================================
# HELPER FUNCTIONS
# =============================================================================================


def stddev_away(namex, errorx, dx):
    if dx > 0:
        print(f"{namex} differs by {errorx / dx:.3f} standard deviations from analytical")
    else:
        print(f"{namex} differs by an undefined number of standard deviations")


def get_analytical(beta, K, O, observables):
    # For a harmonic oscillator with spring constant K,
    # x ~ Normal(x_0, sigma^2), where sigma = 1/sqrt(beta K)

    # Compute the absolute dimensionless free energies of each oscillator analytically.
    # f = - ln(sqrt((2 pi)/(beta K)) )
    print("Computing dimensionless free energies analytically...")

    sigma = (beta * K) ** -0.5
    f_k_analytical = -np.log(np.sqrt(2 * np.pi) * sigma)

    Delta_f_ij_analytical = f_k_analytical - np.vstack(f_k_analytical)

    A_k_analytical = dict()
    A_ij_analytical = dict()

    for observe in observables:
        if observe == "RMS displacement":
            # mean square displacement
            A_k_analytical[observe] = sigma
        if observe == "potential energy":
            # By equipartition
            A_k_analytical[observe] = 1 / (2 * beta) * np.ones(len(K), float)
        if observe == "position":
            # observable is the position
            A_k_analytical[observe] = O
        if observe == "position^2":
            # observable is the position^2
            A_k_analytical[observe] = (1 + beta * K * O**2) / (beta * K)

        A_ij_analytical[observe] = A_k_analytical[observe] - np.vstack(A_k_analytical[observe])

    return f_k_analytical, Delta_f_ij_analytical, A_k_analytical, A_ij_analytical


# =============================================================================================
# PARAMETERS
# =============================================================================================


K_k = np.array([25, 16, 9, 4, 1, 1])  # spring constants for each state
O_k = np.array([0, 1, 2, 3, 4, 5])  # offsets for spring constants
# number of samples from each state (can be zero for some states)
N_k = 10 * np.array([1000, 1000, 1000, 1000, 0, 1000])
Nk_ne_zero = N_k != 0
beta = 1.0  # inverse temperature for all simulations
K_extra = np.array([20, 12, 6, 2, 1])
O_extra = np.array([0.5, 1.5, 2.5, 3.5, 4.5])
observables = ["position", "position^2", "potential energy", "RMS displacement"]

seed = None
# Uncomment the following line to seed the random number generated to
# produce reproducible output.
seed = 0
np.random.seed(seed)

# =============================================================================================
# MAIN
# =============================================================================================

# Determine number of simulations.
K = np.size(N_k)
if np.shape(K_k) != np.shape(N_k):
    msg = f"K_k ({np.shape(K_k):d}) and N_k ({np.shape(N_k):d}) must have same dimensions."
    raise ParameterError(msg)
if np.shape(O_k) != np.shape(N_k):
    msg = f"O_k ({np.shape(K_k):d}) and N_k ({np.shape(N_k):d}) must have same dimensions."
    raise ParameterError(msg)

# Determine maximum number of samples to be drawn for any state.
N_max = np.max(N_k)

f_k_analytical, Delta_f_ij_analytical, A_k_analytical, A_ij_analytical = get_analytical(
    beta, K_k, O_k, observables
)

print(f"This script will draw samples from {K:d} harmonic oscillators.")
print("The harmonic oscillators have equilibrium positions:", O_k)
print("and spring constants:", K_k)
print(
    "and the following number of samples will be drawn from each",
    "(can be zero if no samples drawn):",
    N_k,
)


# =============================================================================================
# Generate independent data samples from K one-dimensional harmonic oscillators centered at q = 0.
# =============================================================================================

print("generating samples...")
randomsample = testsystems.harmonic_oscillators.HarmonicOscillatorsTestCase(
    O_k=O_k, K_k=K_k, beta=beta
)
x_kn, u_kln, N_k = randomsample.sample(N_k, mode="u_kln", seed=seed)

# get the unreduced energies
U_kln = u_kln / beta

# =============================================================================================
# Estimate free energies and expectations.
# =============================================================================================

print("======================================")
print("      Initializing MBAR               ")
print("======================================")

# Estimate free energies from simulation using MBAR.
print("Estimating relative free energies from simulation (this may take a while)...")

# Initialize the MBAR class, determining the free energies.
mbar = MBAR(u_kln, N_k, relative_tolerance=1.0e-10, verbose=True)
# Get matrix of dimensionless free energy differences and uncertainty estimate.

print("=============================================")
print("      Testing compute_free_energy_differences       ")
print("=============================================")

results = mbar.compute_free_energy_differences()
Delta_f_ij_estimated = results["Delta_f"]
dDelta_f_ij_estimated = results["dDelta_f"]

# Compute error from analytical free energy differences.
Delta_f_ij_error = Delta_f_ij_estimated - Delta_f_ij_analytical

print("Error in free energies is:")
print(Delta_f_ij_error)
print("Uncertainty in free energies is:")
print(dDelta_f_ij_estimated)

print("Standard deviations away is:")
# mathematical manipulation to avoid dividing by zero errors; we don't care
# about the diagnonals, since they are identically zero.
df_ij_mod = dDelta_f_ij_estimated + np.identity(K)
stdevs = np.abs(Delta_f_ij_error / df_ij_mod)
for k in range(K):
    stdevs[k, k] = 0
print(stdevs)

print("==============================================")
print("             Testing computeBAR               ")
print("==============================================")

nonzero_indices = np.arange(K)[Nk_ne_zero]
Knon = len(nonzero_indices)
for i in range(Knon - 1):
    k = nonzero_indices[i]
    k1 = nonzero_indices[i + 1]
    w_F = u_kln[k, k1, 0 : N_k[k]] - u_kln[k, k, 0 : N_k[k]]  # forward work
    w_R = u_kln[k1, k, 0 : N_k[k1]] - u_kln[k1, k1, 0 : N_k[k1]]  # reverse work
    results = bar(w_F, w_R)
    df_bar = results["Delta_f"]
    ddf_bar = results["dDelta_f"]
    bar_analytical = f_k_analytical[k1] - f_k_analytical[k]
    bar_error = bar_analytical - df_bar
    print(
        f"BAR estimator for reduced free energy from states {k:d} to {k1:d} is {df_bar:f} +/- {ddf_bar:f}"
    )
    stddev_away("BAR estimator", bar_error, ddf_bar)

print("==============================================")
print("             Testing EXP               ")
print("==============================================")

print("EXP forward free energy")
for k in range(K - 1):
    if N_k[k] != 0:
        # forward work
        w_F = u_kln[k, k + 1, 0 : N_k[k]] - u_kln[k, k, 0 : N_k[k]]
        results = exp(w_F)
        df_exp = results["Delta_f"]
        ddf_exp = results["dDelta_f"]
        exp_analytical = f_k_analytical[k + 1] - f_k_analytical[k]
        exp_error = exp_analytical - df_exp
        print(f"df from states {k:d} to {k + 1:d} is {df_exp:f} +/- {ddf_exp:f}")
        stddev_away("df", exp_error, ddf_exp)

print("EXP reverse free energy")
for k in range(1, K):
    if N_k[k] != 0:
        w_R = u_kln[k, k - 1, 0 : N_k[k]] - u_kln[k, k, 0 : N_k[k]]  # reverse work
        results = exp(w_R)
        df_exp = -results["Delta_f"]
        ddf_exp = results["dDelta_f"]
        exp_analytical = f_k_analytical[k] - f_k_analytical[k - 1]
        exp_error = exp_analytical - df_exp
        print(f"df from states {k:d} to {k - 1:d} is {df_exp:f} +/- {ddf_exp:f}")
        stddev_away("df", exp_error, ddf_exp)

print("==============================================")
print("             Testing computeGauss               ")
print("==============================================")

print("Gaussian forward estimate")
for k in range(K - 1):
    if N_k[k] != 0:
        w_F = u_kln[k, k + 1, 0 : N_k[k]] - u_kln[k, k, 0 : N_k[k]]  # forward work
        results = exp_gauss(w_F)
        df_gauss = results["Delta_f"]
        ddf_gauss = results["dDelta_f"]
        gauss_analytical = f_k_analytical[k + 1] - f_k_analytical[k]
        gauss_error = gauss_analytical - df_gauss
        print(
            f"df for reduced free energy from states {k:d} to {k + 1:d} is {df_gauss:f} +/- {ddf_gauss:f}"
        )
        stddev_away("df", gauss_error, ddf_gauss)

print("Gaussian reverse estimate")
for k in range(1, K):
    if N_k[k] != 0:
        # reverse work
        w_R = u_kln[k, k - 1, 0 : N_k[k]] - u_kln[k, k, 0 : N_k[k]]
        results = exp_gauss(w_R)
        df_gauss = results["Delta_f"]
        ddf_gauss = results["dDelta_f"]
        gauss_analytical = f_k_analytical[k] - f_k_analytical[k - 1]
        gauss_error = gauss_analytical - df_gauss
        print(
            f"df for reduced free energy from states {k:d} to {k - 1:d} is {df_gauss:f} +/- {ddf_gauss:f}"
        )
        stddev_away("df", gauss_error, ddf_gauss)

print("======================================")
print("      Testing compute_expectations")
print("======================================")

A_kn_all = dict()
A_k_estimated_all = dict()
A_kl_estimated_all = dict()
N = np.sum(N_k)

for observe in observables:
    print("============================================")
    print(f"      Testing observable '{observe}'")
    print("============================================")

    if observe == "RMS displacement":
        state_dependent = True
        A_kn = np.zeros([K, N], dtype=np.float64)
        n = 0
        for k in range(K):
            for nk in range(N_k[k]):
                # observable is the squared displacement
                A_kn[:, n] = (x_kn[k, nk] - O_k[:]) ** 2
                n += 1

    # observable is the potential energy, a 3D array since the
    # potential energy is a function of the thermodynamic state
    elif observe == "potential energy":
        state_dependent = True
        A_kn = np.zeros([K, N], dtype=np.float64)
        n = 0
        for k in range(0, K):
            for nk in range(0, N_k[k]):
                A_kn[:, n] = U_kln[k, :, nk]
                n += 1

    # observable for estimation is the position
    elif observe == "position":
        state_dependent = False
        A_kn = np.zeros([K, N_max], dtype=np.float64)
        for k in range(0, K):
            A_kn[k, 0 : N_k[k]] = x_kn[k, 0 : N_k[k]]

    # observable for estimation is the position^2
    elif observe == "position^2":
        state_dependent = False
        A_kn = np.zeros([K, N_max], dtype=np.float64)
        for k in range(0, K):
            A_kn[k, 0 : N_k[k]] = x_kn[k, 0 : N_k[k]] ** 2

    results = mbar.compute_expectations(A_kn, state_dependent=state_dependent)
    A_k_estimated = results["mu"]
    dA_k_estimated = results["sigma"]

    # need to additionally transform to get the square root
    if observe == "RMS displacement":
        A_k_estimated = np.sqrt(A_k_estimated)
        # Compute error from analytical observable estimate.
        dA_k_estimated = dA_k_estimated / (2 * A_k_estimated)

    As_k_estimated = np.zeros([K], np.float64)
    dAs_k_estimated = np.zeros([K], np.float64)

    # 'standard' expectation averages - not defined if no samples
    nonzeros = np.arange(K)[Nk_ne_zero]

    totaln = 0
    for k in nonzeros:
        if (observe == "position") or (observe == "position^2"):
            As_k_estimated[k] = np.average(A_kn[k, 0 : N_k[k]])
            dAs_k_estimated[k] = np.sqrt(np.var(A_kn[k, 0 : N_k[k]]) / (N_k[k] - 1))
        elif (observe == "RMS displacement") or (observe == "potential energy"):
            totalp = totaln + N_k[k]
            As_k_estimated[k] = np.average(A_kn[k, totaln:totalp])
            dAs_k_estimated[k] = np.sqrt(np.var(A_kn[k, totaln:totalp]) / (N_k[k] - 1))
            totaln = totalp
            if observe == "RMS displacement":
                As_k_estimated[k] = np.sqrt(As_k_estimated[k])
                dAs_k_estimated[k] = dAs_k_estimated[k] / (2 * As_k_estimated[k])

    A_k_error = A_k_estimated - A_k_analytical[observe]
    As_k_error = As_k_estimated - A_k_analytical[observe]

    print("------------------------------")
    print("Now testing 'averages' mode")
    print("------------------------------")

    print(f"Analytical estimator of {observe} is")
    print(A_k_analytical[observe])

    print(f"MBAR estimator of the {observe} is")
    print(A_k_estimated)

    print("MBAR estimators differ by X standard deviations")
    stdevs = np.abs(A_k_error / dA_k_estimated)
    print(stdevs)

    print(f"Standard estimator of {observe} is (states with samples):")
    print(As_k_estimated[Nk_ne_zero])

    print("Standard estimators differ by X standard deviations (states with samples)")
    stdevs = np.abs(As_k_error[Nk_ne_zero] / dAs_k_estimated[Nk_ne_zero])
    print(stdevs)

    results = mbar.compute_expectations(
        A_kn, state_dependent=state_dependent, output="differences"
    )
    A_kl_estimated = results["mu"]
    dA_kl_estimated = results["sigma"]

    print("------------------------------")
    print("Now testing 'differences' mode")
    print("------------------------------")

    if (
        "RMS displacement" != observe
    ):  # can't test this, because we're actually computing the expectation of
        # the mean square displacement, and so the differences are <a_i^2> - <a_j^2>,
        # not sqrt<a_i>^2 - sqrt<a_j>^2
        A_kl_analytical = A_k_analytical[observe] - np.vstack(A_k_analytical[observe])
        A_kl_error = A_kl_estimated - A_kl_analytical

        print(f"Analytical estimator of differences of {observe} is")
        print(A_kl_analytical)

        print(f"MBAR estimator of the differences of {observe} is")
        print(A_kl_estimated)

        print("MBAR estimators differ by X standard deviations")
        stdevs = np.abs(A_kl_error / (dA_kl_estimated + np.identity(K)))
        for k in range(K):
            stdevs[k, k] = 0
        print(stdevs)

    # save up the A_k for use in compute_multiple_expectations
    A_kn_all[observe] = A_kn
    A_k_estimated_all[observe] = A_k_estimated
    A_kl_estimated_all[observe] = A_kl_estimated

print("=============================================")
print("      Testing compute_multiple_expectations")
print("=============================================")

# have to exclude the potential and RMS displacemet for now, not functions
# of a single state
observables_single = ["position", "position^2"]

A_ikn = np.zeros([len(observables_single), K, N_k.max()], np.float64)
for i, observe in enumerate(observables_single):
    A_ikn[i, :, :] = A_kn_all[observe]
for i in range(K):
    results = mbar.compute_multiple_expectations(A_ikn, u_kln[:, i, :], compute_covariance=True)
    A_i = results["mu"]
    dA_ij = results["sigma"]
    Ca_ij = results["covariances"]
    print(f"Averages for state {i:d}")
    print(A_i)
    print(f"Uncertainties for state {i:d}")
    print(dA_ij)
    print(f"Correlation matrix between observables for state {i:d}")
    print(Ca_ij)

print("============================================")
print("      Testing compute_entropy_and_enthalpy")
print("============================================")

results = mbar.compute_entropy_and_enthalpy(u_kn=u_kln, verbose=True)
Delta_f_ij = results["Delta_f"]
dDelta_f_ij = results["dDelta_f"]
Delta_u_ij = results["Delta_u"]
dDelta_u_ij = results["dDelta_u"]
Delta_s_ij = results["Delta_s"]
dDelta_s_ij = results["dDelta_s"]

print("Free energies")
print(Delta_f_ij)
print(dDelta_f_ij)
diffs1 = Delta_f_ij - Delta_f_ij_estimated
print(
    f"maximum difference between values computed here and in computeFreeEnergies is {np.max(diffs1):g}"
)
if np.max(np.abs(diffs1)) > 1.0e-10:
    print("Difference in values from computeFreeEnergies")
    print(diffs1)
diffs2 = dDelta_f_ij - dDelta_f_ij_estimated
print(
    f"maximum difference between uncertainties computed here and in computeFreeEnergies is {np.max(diffs2):g}"
)
if np.max(np.abs(diffs2)) > 1.0e-10:
    print("Difference in expectations from computeFreeEnergies")
    print(diffs2)

print("Energies")
print(Delta_u_ij)
print(dDelta_u_ij)
U_k = A_k_estimated_all["potential energy"]
expectations = U_k - np.vstack(U_k)
diffs1 = Delta_u_ij - expectations
print(
    f"maximum difference between values computed here and in compute_expectations is {np.max(diffs1):g}"
)
if np.max(np.abs(diffs1)) > 1.0e-10:
    print("Difference in values from compute_expectations")
    print(diffs1)

print("Entropies")
print(Delta_s_ij)
print(dDelta_s_ij)

# analytical entropy estimate
s_k_analytical = 0.5 / beta - f_k_analytical
Delta_s_ij_analytical = s_k_analytical - np.vstack(s_k_analytical)

Delta_s_ij_error = Delta_s_ij_analytical - Delta_s_ij
print("Error in entropies is:")
print(Delta_f_ij_error)

print("Standard deviations away is:")
# mathematical manipulation to avoid dividing by zero errors; we don't care
# about the diagnonals, since they are identically zero.
ds_ij_mod = dDelta_s_ij + np.identity(K)
stdevs = np.abs(Delta_s_ij_error / ds_ij_mod)
for k in range(K):
    stdevs[k, k] = 0
print(stdevs)

print("============================================")
print("      Testing compute_perturbed_free_energies")
print("============================================")

L = np.size(K_extra)
f_k_analytical, Delta_f_ij_analytical, A_k_analytical, A_ij_analytical = get_analytical(
    beta, K_extra, O_extra, observables
)

if np.size(O_extra) != np.size(K_extra):
    raise ParameterError(
        f"O_extra ({np.shape(K_k):d}) and K_extra ({np.shape(N_k):d}) must have the same dimensions."
    )

unew_kln = np.zeros([K, L, np.max(N_k)], np.float64)
for k in range(K):
    for l in range(L):
        unew_kln[k, l, 0 : N_k[k]] = (K_extra[l] / 2.0) * (x_kn[k, 0 : N_k[k]] - O_extra[l]) ** 2

results = mbar.compute_perturbed_free_energies(unew_kln)
Delta_f_ij_estimated = results["Delta_f"]
dDelta_f_ij_estimated = results["dDelta_f"]

Delta_f_ij_error = Delta_f_ij_estimated - Delta_f_ij_analytical

print("Error in free energies is:")
print(Delta_f_ij_error)

print("Standard deviations away is:")
# mathematical manipulation to avoid dividing by zero errors; we don't care
# about the diagnonals, since they are identically zero.
df_ij_mod = dDelta_f_ij_estimated + np.identity(L)
stdevs = np.abs(Delta_f_ij_error / df_ij_mod)
for l in range(L):
    stdevs[l, l] = 0
print(stdevs)

print("============================================")
print("      Testing compute_expectation (new states)")
print("============================================")

nth = 3
# test the nth "extra" states, O_extra[nth] & K_extra[nth]
for observe in observables:
    print("============================================")
    print(f"      Testing observable '{observe}'")
    print("============================================")

    if observe == "RMS displacement":
        state_dependent = True
        A_kn = np.zeros([K, 1, N_max], dtype=np.float64)
        for k in range(0, K):
            # observable is the squared displacement
            A_kn[k, 0, 0 : N_k[k]] = (x_kn[k, 0 : N_k[k]] - O_extra[nth]) ** 2

    # observable is the potential energy, a 3D array since the potential energy is a function of
    # thermodynamic state
    elif observe == "potential energy":
        state_dependent = True
        A_kn = unew_kln[:, [nth], :] / beta

    # position and position^2 can use the same observables
    # observable for estimation is the position
    elif observe == "position":
        state_dependent = False
        A_kn = A_kn_all["position"]

    elif observe == "position^2":
        state_dependent = False
        A_kn = A_kn_all["position^2"]

    A_k_estimated, dA_k_estimated
    results = mbar.compute_expectations(
        A_kn, unew_kln[:, [nth], :], state_dependent=state_dependent
    )
    A_k_estimated = results["mu"]
    dA_k_estimated = results["sigma"]
    # need to additionally transform to get the square root
    if observe == "RMS displacement":
        A_k_estimated = np.sqrt(A_k_estimated)
        dA_k_estimated = dA_k_estimated / (2 * A_k_estimated)

    A_k_error = A_k_estimated - A_k_analytical[observe][nth]

    print(f"Analytical estimator of {observe} is")
    print(A_k_analytical[observe][nth])

    print(f"MBAR estimator of the {observe} is")
    print(A_k_estimated)

    print("MBAR estimators differ by X standard deviations")
    stdevs = np.abs(A_k_error / dA_k_estimated)
    print(stdevs)

print("============================================")
print("      Testing compute_overlap")
print("============================================")

results = mbar.compute_overlap()
O = results["scalar"]
O_i = results["eigenvalues"]
O_ij = results["matrix"]

print("Overlap matrix output")
print(O_ij)

for k in range(K):
    print(f"Sum of row {k:d} is {np.sum(O_ij[k, :]):f} (should be 1),", end=" ")
    if np.abs(np.sum(O_ij[k, :]) - 1) < 1.0e-10:
        print("looks like it is.")
    else:
        print("but it's not.")

print("Eigenvalues of overlap matrix:")
print(O_i)


print("Overlap scalar measure: (1-lambda_2)")
print(O)

print("============================================")
print("    Testing compute_effective_sample_number")
print("============================================")

N_eff = mbar.compute_effective_sample_number(verbose=True)
print("Effective Sample number")
print(N_eff)
print("Compare stanadrd estimate of <x> with the MBAR estimate of <x>")
print("We should have that with MBAR, err_MBAR = sqrt(N_k/N_eff)*err_standard,")
print("so standard (scaled) results should be very close to MBAR results.")
print("No standard estimate exists for states that are not sampled.")
A_kn = x_kn
results = mbar.compute_expectations(A_kn)
val_mbar = results["mu"]
err_mbar = results["sigma"]
err_standard = np.zeros([K], dtype=np.float64)
err_scaled = np.zeros([K], dtype=np.float64)

for k in range(K):
    if N_k[k] != 0:
        # use position
        err_standard[k] = np.std(A_kn[k, 0 : N_k[k]]) / np.sqrt(N_k[k] - 1)
        err_scaled[k] = np.std(A_kn[k, 0 : N_k[k]]) / np.sqrt(N_eff[k] - 1)

print("                   ", end=" ")
for k in range(K):
    print(f"     {k:d}    ", end=" ")
print("")
print("MBAR             :", end=" ")
print(err_mbar)
print("standard         :", end=" ")
print(err_standard)
print("sqrt N_k/N_eff   :", end=" ")
print(np.sqrt(N_k / N_eff))
print("Standard (scaled):", end=" ")
print(err_standard * np.sqrt(N_k / N_eff))

print("============================================")
print("      Testing free energy surface functions   ")
print("============================================")

# For 2-D, The equilibrium distribution is given analytically by
#   p(x;beta,K) = sqrt[(beta K) / (2 pi)] exp[-beta K [(x-mu)^2] / 2]
#
# The dimensionless free energy is therefore
#   f(beta,K) = - (1/2) * ln[ (2 pi) / (beta K) ]
#
# In this problem, we are investigating the sum of two Gaussians, once
# centered at 0, and others centered at grid points.
#
#   V(x;K) = (K0/2) * [(x-x_0)^2]
#
# For 1-D, The equilibrium distribution is given analytically by
#   p(x;beta,K) = 1/N exp[-beta (K0 [x^2] / 2  + KU [(x-mu)^2] / 2)]
#   Where N is the normalization constant.
#
# The dimensionless free energy is the integral of this, and can be computed as:
#   f(beta,K)           = - ln [ (2*np.pi/(Ko+Ku))^(d/2) exp[ -Ku*Ko mu' mu / 2(Ko +Ku)]
#   f(beta,K) - fzero   = -Ku*Ko / 2(Ko+Ku)  = 1/(1/(Ku/2) + 1/(K0/2))


def generate_fes_data(
    ndim=1, nbinsperdim=15, nsamples=1000, K0=20.0, Ku=100.0, gridscale=0.2, xrange=((-3, 3),)
):
    x0 = np.zeros([ndim], np.float64)  # center of base potential
    numbrellas = 1
    nperdim = np.zeros([ndim], int)
    for d in range(ndim):
        nperdim[d] = xrange[d][1] - xrange[d][0] + 1
        numbrellas *= nperdim[d]

    print(f"There are a total of {numbrellas:d} umbrellas.")

    # Enumerate umbrella centers, and compute the analytical free energy of
    # that umbrella
    print("Constructing umbrellas...")
    ksum = (Ku + K0) / beta
    kprod = (Ku * K0) / (beta * beta)
    f_k_analytical = np.zeros(numbrellas, np.float64)
    # xu_i[i,:] is the center of umbrella i
    xu_i = np.zeros([numbrellas, ndim], np.float64)

    dp = np.zeros(ndim, int)
    dp[0] = 1
    for d in range(1, ndim):
        dp[d] = nperdim[d] * dp[d - 1]

    umbrella_zero = 0
    for i in range(numbrellas):
        center = []
        for d in range(ndim):
            val = gridscale * ((int(i // dp[d])) % nperdim[d] + xrange[d][0])
            center.append(val)
        center = np.array(center)
        xu_i[i, :] = center
        mu2 = np.dot(center, center)
        f_k_analytical[i] = np.log(
            (ndim * np.pi / ksum) ** (3.0 / 2.0) * np.exp(-kprod * mu2 / (2.0 * ksum))
        )
        # assumes that we have one state that is at the zero.
        if np.all(center == 0.0):
            umbrella_zero = i
        i += 1
        f_k_analytical -= f_k_analytical[umbrella_zero]

    print(f"Generating {nsamples:d} samples for each of {numbrellas:d} umbrellas...")
    x_n = np.zeros([numbrellas * nsamples, ndim], np.float64)

    for i in range(numbrellas):
        for dim in range(ndim):
            # Compute mu and sigma for this dimension for sampling from V0(x) + Vu(x).
            # Product of Gaussians: N(x ; a, A) N(x ; b, B) = N(a ; b , A+B) x N(x ; c, C) where
            # C = 1/(1/A + 1/B)
            # c = C(a/A+b/B)
            # A = 1/K0, B = 1/Ku
            sigma = 1.0 / (K0 + Ku)
            mu = sigma * (x0[dim] * K0 + xu_i[i, dim] * Ku)
            # Generate normal deviates for this dimension.
            x_n[i * nsamples : (i + 1) * nsamples, dim] = np.random.normal(
                mu, np.sqrt(sigma), [nsamples]
            )

    u_kn = np.zeros([numbrellas, nsamples * numbrellas], np.float64)
    # Compute reduced potential due to V0.
    u_n = beta * (K0 / 2) * np.sum((x_n[:, :] - x0) ** 2, axis=1)
    for k in range(numbrellas):
        # reduced potential due to umbrella k
        uu = beta * (Ku / 2) * np.sum((x_n[:, :] - xu_i[k, :]) ** 2, axis=1)
        u_kn[k, :] = u_n + uu

    return u_kn, u_n, x_n, f_k_analytical


nbinsperdim = 15
gridscale = 0.2
nsamples = 1000
ndim = 1
K0 = 20.0
Ku = 100.0
print("============================================")
print("      Test 1: 1D free energy profile   ")
print("============================================")

xrange = [[-3, 3]]
ndim = 1
u_kn, u_n, x_n, f_k_analytical = generate_fes_data(
    K0=K0,
    Ku=Ku,
    ndim=ndim,
    nbinsperdim=nbinsperdim,
    nsamples=nsamples,
    gridscale=gridscale,
    xrange=xrange,
)
numbrellas = (np.shape(u_kn))[0]
N_k = nsamples * np.ones([numbrellas], int)
print("Solving for free energies of state ...")
mbar = MBAR(u_kn, N_k)

# Histogram bins are indexed using the scheme:
# index = 1 + np.floor((x[0] - xmin)/dx) + nbins*np.floor((x[1] - xmin)/dy)
# index = 0 is reserved for samples outside of the allowed domain
xmin = gridscale * (np.min(xrange[0][0]) - 1 / 2.0)
xmax = gridscale * (np.max(xrange[0][1]) + 1 / 2.0)
dx = (xmax - xmin) / nbinsperdim
nbins = 1 + nbinsperdim**ndim
bin_edges = np.linspace(xmin, xmax, nbins)  # list of bin edges.
bin_centers = np.zeros([nbins, ndim], np.float64)

ibin = 1
fes_analytical = np.zeros([nbins], np.float64)
minmu2 = 1000000
zeroindex = 0
# construct the bins and the fes
for i in range(nbinsperdim):
    xbin = xmin + dx * (i + 0.5)
    bin_centers[ibin, 0] = xbin
    mu2 = xbin * xbin
    if mu2 < minmu2:
        minmu2 = mu2
        zeroindex = ibin
    fes_analytical[ibin] = K0 * mu2 / 2.0
    ibin += 1
fzero = fes_analytical[zeroindex]
fes_analytical -= fzero
fes_analytical[0] = 0

bin_n = np.zeros([numbrellas * nsamples], int)
# Determine indices of those within bounds.
within_bounds = (x_n[:, 0] >= xmin) & (x_n[:, 0] < xmax)
# Determine states for these.
bin_n[within_bounds] = 1 + np.floor((x_n[within_bounds, 0] - xmin) / dx)
# Determine indices of bins that are not empty.
bin_counts = np.zeros([nbins], int)
for i in range(nbins):
    bin_counts[i] = (bin_n == i).sum()

# Compute fre energy profile, first with histograms
print("Solving for free energies of state to initialize free energy profile...")
mbar_options = dict()
mbar_options["verbose"] = False
fes = FES(u_kn, N_k, mbar_options=mbar_options)
print("Computing free energy profile ...")
histogram_parameters = dict()
histogram_parameters["bin_edges"] = bin_edges
fes.generate_fes(u_n, x_n, histogram_parameters=histogram_parameters)
results = fes.get_fes(
    bin_centers[:, 0],
    reference_point="from-specified",
    fes_reference=0.0,
    uncertainty_method="analytical",
)
f_ih = results["f_i"]
df_ih = results["df_i"]

# now estimate the PDF with a kde
kde_parameters = dict()
kde_parameters["bandwidth"] = dx / 3.0
fes.generate_fes(u_n, x_n, fes_type="kde", n_bootstraps=20, kde_parameters=kde_parameters)
results_kde = fes.get_fes(
    bin_centers,
    reference_point="from-specified",
    fes_reference=0.0,
    uncertainty_method="bootstrap",
)
f_ik = results_kde["f_i"]
df_ik = results_kde["df_i"]

# Show free energy and uncertainty of each occupied bin relative to lowest
# free energy

print("1D free energy profile:")
print(f"{bin_counts[0]:d} counts out of {numbrellas * nsamples:d} counts not in any bin")
print(
    f"{'bin':>8s} {'x':>6s} {'N':>8s} {'true':>10s}"
    f"{'f_hist':>10s} {'err_hist':>10s} {'df_hist':>10s} {'sig_hist':>8s}"
    f"{'f_kde':>10s} {'err_kde':>10s} {'df_kde':>10s} {'sig_kde':>8s}"
)

for i in range(1, nbins):
    error_h = fes_analytical[i] - f_ih[i]
    error_k = fes_analytical[i] - f_ik[i]
    if df_ih[i] > 0:
        stdevs_h = np.abs(error_h) / df_ih[i]
    else:
        stdevs_h = 0

    if df_ik[i] > 0:
        stdevs_k = np.abs(error_k) / df_ik[i]
    else:
        stdevs_k = 0

    print(
        f"{i:>8d} {bin_centers[i, 0]:>6.2f} {bin_counts[i]:>8d} {fes_analytical[i]:>10.3f}"
        f"{f_ih[i]:>10.3f} {error_h:>10.3f} {df_ih[i]:>10.3f} {stdevs_h:>8.2f}"
        f"{f_ik[i]:>10.3f} {error_k:>10.3f} {df_ik[i]:>10.3f} {stdevs_k:>8.2f}"
    )

print("============================================")
print("      Test 2: 2D free energy surface  ")
print("============================================")

xrange = [[-3, 3], [-3, 3]]
ndim = 2
nsamples = 500
u_kn, u_n, x_n, f_k_analytical = generate_fes_data(
    K0=K0,
    Ku=Ku,
    ndim=ndim,
    nbinsperdim=nbinsperdim,
    nsamples=nsamples,
    gridscale=gridscale,
    xrange=xrange,
)
numbrellas = np.shape(u_kn)[0]
N_k = nsamples * np.ones([numbrellas], int)
print("Solving for free energies of state ...")
mbar = MBAR(u_kn, N_k)

# The dimensionless free energy is the integral of this, and can be computed as:
#   f(beta,K)           = - ln [ (2*np.pi/(Ko+Ku))^(d/2) exp[ -Ku*Ko mu' mu / 2(Ko +Ku)]
#   f(beta,K) - fzero   = -Ku*Ko / 2(Ko+Ku)  = 1/(1/(Ku/2) + 1/(K0/2))
# for computing harmonic samples

# Can compare the free energies computed with MBAR if desired with
# f_k_analytical

# Histogram bins are indexed using the scheme:
# index = 1 + np.floor((x[0] - xmin)/dx) + nbins*np.floor((x[1] - xmin)/dy)
# index = 0 is reserved for samples outside of the allowed domain

xmin = gridscale * (np.min(xrange[0][0]) - 1 / 2.0)
xmax = gridscale * (np.max(xrange[0][1]) + 1 / 2.0)
ymin = gridscale * (np.min(xrange[1][0]) - 1 / 2.0)
ymax = gridscale * (np.max(xrange[1][1]) + 1 / 2.0)
dx = (xmax - xmin) / nbinsperdim
dy = (ymax - ymin) / nbinsperdim
nbins = 1 + nbinsperdim**ndim
bin_centers = np.zeros([nbins, ndim], np.float64)

ibin = 1  # first reserved for something outside.
fes_analytical = np.zeros([nbins], np.float64)
minmu2 = 1000000
zeroindex = 0
# construct the bins and the fes
for i in range(nbinsperdim):
    xbin = xmin + dx * (i + 0.5)
    for j in range(nbinsperdim):
        # Determine (x,y) of bin center.
        ybin = ymin + dy * (j + 0.5)
        bin_centers[ibin, 0] = xbin
        bin_centers[ibin, 1] = ybin
        mu2 = xbin * xbin + ybin * ybin
        if mu2 < minmu2:
            minmu2 = mu2
            zeroindex = ibin
        fes_analytical[ibin] = K0 * mu2 / 2.0
        ibin += 1
fzero = fes_analytical[zeroindex]
fes_analytical -= fzero

bin_n = np.zeros([numbrellas * nsamples], int)
# Determine indices of those within bounds.
within_bounds = (x_n[:, 0] >= xmin) & (x_n[:, 0] < xmax) & (x_n[:, 1] >= ymin) & (x_n[:, 1] < ymax)
# Determine states for these.
xgrid = (x_n[within_bounds, 0] - xmin) / dx
ygrid = (x_n[within_bounds, 1] - ymin) / dy
bin_n[within_bounds] = 1 + xgrid.astype(int) + nbinsperdim * ygrid.astype(int)

# Determine indices of bins that are not empty.
bin_counts = np.zeros([nbins], int)
for i in range(nbins):
    bin_counts[i] = (bin_n == i).sum()

# Compute free energy surface, first using histograms weighted with MBAR
print("Computing free energy surface ...")
fes = FES(u_kn, N_k)

# for 2D bins, we input a list of the bin edges in each dimension.
histogram_parameters["bin_edges"] = [
    np.linspace(xmin, xmax, nbinsperdim + 1),
    np.linspace(ymin, ymax, nbinsperdim + 1),
]  # list of histogram edges.
fes.generate_fes(u_n, x_n, fes_type="histogram", histogram_parameters=histogram_parameters)
delta = 0.0001  # to break ties in things being too close.

results = fes.get_fes(
    bin_centers + delta,
    reference_point="from-specified",
    fes_reference=[0, 0],
    uncertainty_method="analytical",
)
f_i = results["f_i"]
df_i = results["df_i"]

# now generate the kernel density estimate
kde_parameters["bandwidth"] = 0.5 * dx
fes.generate_fes(u_n, x_n, fes_type="kde", kde_parameters=kde_parameters)
results_kde = fes.get_fes(bin_centers, reference_point="from-specified", fes_reference=[0, 0])
f_ik = results_kde["f_i"]

# Show free energy and uncertainty of each occupied bin relative to lowest
# free energy
print("2D FES:")

print(f"{bin_counts[0]:d} counts out of {numbrellas * nsamples:d} counts not in any bin")
print("Uncertainties only calculated for histogram methods")
print(
    f"{'bin':>8s} {'x':>6s} {'y':>6s} {'N':>8s} {'f_hist':>10s} {'f_kde':>10s} {'true':>10s} {'err_hist':>10s} {'err_kde':>10s} {'df':>10s} {'sigmas':>8s}"
)
for i in range(1, nbins):
    if i == zeroindex:
        stdevs = 0
        df_i[0] = 0
    else:
        error = fes_analytical[i] - f_i[i]
        if df_i[i] > 0:
            stdevs = np.abs(error) / df_i[i]
        else:
            stdevs = np.nan
    print(
        f"{i:>8d} "
        f"{bin_centers[i, 0]:>6.2f} "
        f"{bin_centers[i, 1]:>6.2f} "
        f"{bin_counts[i]:>8d} "
        f"{f_i[i]:>10.3f} "
        f"{f_ik[i]:>10.3f} "
        f"{fes_analytical[i]:>10.3f} "
        f"{error:>10.3f} "
        f"{fes_analytical[i]-f_ik[i]:>10.3f} "
        f"{df_i[i]:>10.3f} "
        f"{stdevs:>8.2f}"
    )



##############################################################################
# pymbar: A Python Library for MBAR
#
# Copyright 2016-2017 University of Colorado Boulder
# Copyright 2010-2017 Memorial Sloan-Kettering Cancer Center
# Portions of this software are Copyright 2010-2016 University of Virginia
#
# Authors: Michael Shirts, John Chodera
# Contributors: Kyle Beauchamp
#
# pymbar is free software: you can redistribute it and/or modify
# it under the terms of the MIT License
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# MIT License for more details.
#
# You should have received a copy of the MIT License along with pymbar.
##############################################################################

import logging
from textwrap import dedent
import numpy as np
import scipy
import scipy.special
import scipy.stats


logger = logging.getLogger(__name__)


def order_replicates(replicates, K):
    """
    TODO: Add description for this function and types for parameters

    Parameters
    ----------
    replicates:
        An array of replicates, and the size of the data.

    Returns
    -------
    np.array
        a Nxdims array of the data in the replicates, normalized by the standard deviation
    """

    dims = np.shape(replicates[0]["destimated"])

    sigma = replicates[0]["destimated"]
    zerosigma = sigma == 0
    sigmacorr = (
        zerosigma  # we need to avoid errors with zero standard errors.  We will ignore them later.
    )
    sigma += sigmacorr

    yi = []
    for replicate_index, replicate in enumerate(replicates):
        yi.append(replicate["error"] / sigma)
    yiarray = np.asarray(yi)
    sortedyi = np.zeros(np.shape(yiarray))
    if len(dims) == 0:
        sortedyi[:] = np.sort(yiarray)
    elif len(dims) == 1:
        for i in range(K):
            sortedyi[:, i] = np.sort(yiarray[:, i])
    elif len(dims) == 2:
        for i in range(K):
            for j in range(K):
                sortedyi[:, i, j] = np.sort(yiarray[:, i, j])

    # remove the correction so we have zero sigmas again
    sigma -= sigmacorr
    return sortedyi


def anderson_darling(replicates, K):
    """
    TODO: Description here

    Parameters
    ----------
        replicates: list of replicates
        K: number of replicates

    Returns
    -------
    type
        Anderson-Darling statistics. See:
        http://en.wikipedia.org/wiki/Anderson%E2%80%93Darling_test

    Notes
    -----
    Since both sigma and mu are known (mu exactly, sigma as an estimate from mbar),
    we can apply the case 1 test.

    Because sigma is not precise, we should accept a higher threshold than the 1%
    threshold listed below to throw an error:

        15%  1.610
        10%  1.933
        5%   2.492
        2.5% 3.070
        1%   3.857

    So we choose something like 4.5.  Note that for lower numbers of
    samples, it's more likely.  2000 samples for each of the
    harmonic_oscillators_distributions.py seems to give good
    results.

    For now, the standard deviation we use is the one from the
    _first_ replicate.
    """
    sortedyi = order_replicates(replicates, K)
    zerosigma = replicates[0]["destimated"] == 0  # ignore the ones with zero values of the std

    N = len(replicates)
    dims = np.shape(replicates[0]["destimated"])
    sum = np.zeros(dims)
    for i in range(N):
        cdfi = scipy.stats.norm.cdf(sortedyi[i])
        sum += (2 * i - 1) * np.log(cdfi) + (2 * (N - i) + 1) * np.log(1 - cdfi)
    A2 = -N - sum / N
    A2[zerosigma] = 0
    return A2


def qq_plot(replicates, K, title="Generic Q-Q plot", filename="qq.pdf"):
    """
    TODO: Description here

    Parameters
    ----------
    replicates : list
        TODO: type and description
    K : int
        TODO: type and description
    title : str, optional="Generic Q-Q plot"
        Plot title
    filename : str, optional="qq.pdf"
        Output path to generated PDF
    """

    import matplotlib
    import matplotlib.pyplot as plt

    sortedyi = order_replicates(replicates, K)
    N = len(replicates)
    dim = len(np.shape(replicates[0]["error"]))
    xvals = scipy.stats.norm.ppf((np.arange(0, N) + 0.5) / N)  # inverse pdf

    if dim == 0:
        nplots = 1
    elif dim == 1:
        nplots = K
    elif dim == 2:
        nplots = K * K

    yy = np.zeros([N, nplots])  # pylint: disable=used-before-assignment

    labelij = dict()
    if dim == 0:
        yy[:, 0] = sortedyi[:]
    elif dim == 1:
        nplots = K
        for i in range(K):
            yy[:, i] = sortedyi[:, i]
    elif dim == 2:
        nplots = K * (K - 1)
        k = 0
        for i in range(K):
            for j in range(K):
                if i != j:
                    yy[:, k] = sortedyi[:, i, j]
                    labelij[k] = [i, j]
                    k += 1

    sq = nplots**0.5
    labelsize = 30.0 / sq
    matplotlib.rc("axes", facecolor="#E3E4FA")
    matplotlib.rc("axes", edgecolor="white")
    matplotlib.rc("xtick", labelsize=labelsize)
    matplotlib.rc("ytick", labelsize=labelsize)
    h = int(sq)
    w = h + 1 + 1 * (sq - h > 0.5)
    fig = plt.figure(figsize=(8, 6))
    for i in range(nplots):
        ax = plt.subplot(h, w, i + 1)
        ms = 75.0 / len(yy[:, i])
        ax.plot(xvals, yy[:, i], color="r", ms=ms, marker="o", mec="r")
        ax.plot(xvals, xvals, color="b", ls="-")
        plt.xlim(xvals.min(), xvals.max())
        if dim == 1:
            ax.annotate(
                r"State $\mathrm{%d}$" % (i),
                xy=(0.5, 0.9),
                xycoords=("axes fraction", "axes fraction"),
                xytext=(0, -2),
                size=labelsize,
                textcoords="offset points",
                va="top",
                ha="center",
                color="#151B54",
                bbox=dict(fc="w", ec="none", alpha=0.5),
            )
        if dim == 2:
            ax.annotate(
                r"State $\mathrm{%d-%d}$" % (labelij[i][0], labelij[i][1]),
                xy=(0.5, 0.9),
                xycoords=("axes fraction", "axes fraction"),
                xytext=(0, -2),
                size=labelsize,
                textcoords="offset points",
                va="top",
                ha="center",
                color="#151B54",
                bbox=dict(fc="w", ec="none", alpha=0.5),
            )
    plt.suptitle(title, fontsize=20)
    plt.savefig(filename)
    plt.close(fig)

    return


def generate_confidence_intervals(replicates, K):
    """
    Parameters
    ----------
    replicates: list
        list of replicates
    K: int
        number of replicates

    Returns
    -------
    alpha_values
        TODO: Description and type
    Pobs
        TODO: Description and type
    Plow
        TODO: Description and type
    Phigh
        TODO: Description and type
    dPobs
        TODO: Description and type
    Pnorm
        TODO: Description and type

    Notes
    -----
    Analyze data.

    By Chebyshev's inequality, we should have
      P(error >= alpha sigma) <= 1 / alpha^2
    so that a lower bound will be
      P(error < alpha sigma) > 1 - 1 / alpha^2
    for any real multiplier 'k', where 'sigma' represents the computed uncertainty (as one standard deviation).

    If the error is normal, we should have
      P(error < alpha sigma) = erf(alpha / sqrt(2))
    """

    msg = """
    The uncertainty estimates are tested in this section.
    If the error is normally distributed, the actual error will be less than a
    multiplier 'alpha' times the computed uncertainty 'sigma' a fraction of
    time given by:
    P(error < alpha sigma) = erf(alpha / sqrt(2))
    For example, the true error should be less than 1.0 * sigma
    (one standard deviation) a total of 68% of the time, and
    less than 2.0 * sigma (two standard deviations) 95% of the time.
    The observed fraction of the time that error < alpha sigma, and its
    uncertainty, is given as 'obs' (with uncertainty 'obs err') below.
    This should be compared to the column labeled 'normal'.
    A weak lower bound that holds regardless of how the error is distributed is given
    by Chebyshev's inequality, and is listed as 'cheby' below.
    Uncertainty estimates are tested for both free energy differences and expectations.
    """
    logger.info(dedent(msg[1:]))

    # error bounds
    min_alpha = 0.1
    max_alpha = 4.0
    nalpha = 40
    alpha_values = np.linspace(min_alpha, max_alpha, num=nalpha)
    Pobs = np.zeros([nalpha], dtype=np.float64)
    dPobs = np.zeros([nalpha], dtype=np.float64)
    Plow = np.zeros([nalpha], dtype=np.float64)
    Phigh = np.zeros([nalpha], dtype=np.float64)
    nreplicates = len(replicates)
    dim = len(np.shape(replicates[0]["estimated"]))
    for alpha_index in range(0, nalpha):
        # Get alpha value.
        alpha = alpha_values[alpha_index]
        # Accumulate statistics across replicates
        a = 1.0
        b = 1.0
        # how many dimensions in the data?

        for replicate_index, replicate in enumerate(replicates):
            # Compute fraction of free energy differences where error <= alpha sigma
            # We only count differences where the analytical difference is larger than a cutoff, so that the results will not be limited by machine precision.
            if dim == 0:
                if np.isnan(replicate["error"]) or np.isnan(replicate["destimated"]):
                    logger.warning("replicate {:d}".format(replicate_index))
                    logger.warning("error")
                    logger.warning(replicate["error"])
                    logger.warning("destimated")
                    logger.warning(replicate["destimated"])
                    raise ArithmeticError("Encountered isnan in computation")
                else:
                    if abs(replicate["error"]) <= alpha * replicate["destimated"]:
                        a += 1.0
                    else:
                        b += 1.0

            elif dim == 1:
                for i in range(0, K):
                    if np.isnan(replicate["error"][i]) or np.isnan(replicate["destimated"][i]):
                        logger.warning("replicate {:d}".format(replicate_index))
                        logger.warning("error")
                        logger.warning(replicate["error"])
                        logger.warning("destimated")
                        logger.warning(replicate["destimated"])
                        raise ArithmeticError("Encountered isnan in computation")
                    else:
                        if abs(replicate["error"][i]) <= alpha * replicate["destimated"][i]:
                            a += 1.0
                        else:
                            b += 1.0

            elif dim == 2:
                for i in range(0, K):
                    for j in range(0, i):
                        if np.isnan(replicate["error"][i, j]) or np.isnan(
                            replicate["destimated"][i, j]
                        ):
                            logger.warning("replicate {:d}".format(replicate_index))
                            logger.warning("ij_error")
                            logger.warning(replicate["error"])
                            logger.warning("ij_estimated")
                            logger.warning(replicate["destimated"])
                            raise ArithmeticError("Encountered isnan in computation")
                        else:
                            if (
                                abs(replicate["error"][i, j])
                                <= alpha * replicate["destimated"][i, j]
                            ):
                                a += 1.0
                            else:
                                b += 1.0

        Pobs[alpha_index] = a / (a + b)
        Plow[alpha_index] = scipy.stats.beta.ppf(0.025, a, b)
        Phigh[alpha_index] = scipy.stats.beta.ppf(0.975, a, b)
        dPobs[alpha_index] = np.sqrt(a * b / ((a + b) ** 2 * (a + b + 1)))

    # Write error as a function of sigma.
    logger.info("Error vs. alpha")
    logger.info(
        "{:5s} {:10s} {:10s} {:16s} {:17s}".format("alpha", "cheby", "obs", "obs err", "normal")
    )
    Pnorm = scipy.special.erf(alpha_values / np.sqrt(2.0))
    for alpha_index in range(0, nalpha):
        alpha = alpha_values[alpha_index]
        logger.info(
            "{:5.1f} {:10.6f} {:10.6f} ({:10.6f},{:10.6f}) {:10.6f}".format(
                alpha,
                1.0 - 1.0 / alpha**2,
                Pobs[alpha_index],
                Plow[alpha_index],
                Phigh[alpha_index],
                Pnorm[alpha_index],
            )
        )

    # compute bias, average, etc - do it by replicate, not by bias
    if dim == 0:
        vals = np.zeros([nreplicates], dtype=np.float64)
        vals_error = np.zeros([nreplicates], dtype=np.float64)
        vals_std = np.zeros([nreplicates], dtype=np.float64)
    elif dim == 1:
        vals = np.zeros([nreplicates, K], dtype=np.float64)
        vals_error = np.zeros([nreplicates, K], dtype=np.float64)
        vals_std = np.zeros([nreplicates, K], dtype=np.float64)
    elif dim == 2:
        vals = np.zeros([nreplicates, K, K], dtype=np.float64)
        vals_error = np.zeros([nreplicates, K, K], dtype=np.float64)
        vals_std = np.zeros([nreplicates, K, K], dtype=np.float64)

    rindex = 0
    for replicate in replicates:
        if dim == 0:
            vals[rindex] = replicate["estimated"]
            vals_error[rindex] = replicate["error"]
            vals_std[rindex] = replicate["destimated"]
        elif dim == 1:
            for i in range(0, K):
                vals[rindex, :] = replicate["estimated"]
                vals_error[rindex, :] = replicate["error"]
                vals_std[rindex, :] = replicate["destimated"]
        elif dim == 2:
            for i in range(0, K):
                for j in range(0, i):
                    vals[rindex, :, :] = replicate["estimated"]
                    vals_error[rindex, :, :] = replicate["error"]
                    vals_std[rindex, :, :] = replicate["destimated"]
        rindex += 1

    aveval = np.average(vals, axis=0)
    standarddev = np.std(vals, axis=0)
    bias = np.average(vals_error, axis=0)
    aveerr = np.average(vals_error, axis=0)
    d2 = vals_error**2
    rms_error = (np.average(d2, axis=0)) ** (1.0 / 2.0)
    d2 = vals_std**2
    ave_std = (np.average(d2, axis=0)) ** (1.0 / 2.0)

    # for now, just print out the data at the end for each
    logger.info("")
    logger.info("     i      average    bias      rms_error     stddev  ave_analyt_std")
    logger.info("---------------------------------------------------------------------")
    if dim == 0:
        pave = aveval
        pbias = bias
        prms = rms_error
        pstdev = standarddev
        pavestd = ave_std
    elif dim == 1:
        for i in range(0, K):
            pave = aveval[i]
            pbias = bias[i]
            prms = rms_error[i]
            pstdev = standarddev[i]
            pavestd = ave_std[i]
            logger.info(
                "{:7d} {:10.4f}  {:10.4f}  {:10.4f}  {:10.4f} {:10.4f}".format(
                    i, pave, pbias, prms, pstdev, pavestd
                )
            )
    elif dim == 2:
        for i in range(0, K):
            pave = aveval[0, i]
            pbias = bias[0, i]
            prms = rms_error[0, i]
            pstdev = standarddev[0, i]
            pavestd = ave_std[0, i]
            logger.info(
                "{:7d} {:10.4f}  {:10.4f}  {:10.4f}  {:10.4f} {:10.4f}".format(
                    i, pave, pbias, prms, pstdev, pavestd
                )
            )

    logger.info(
        "Totals: {:10.4f}  {:10.4f}  {:10.4f}  {:10.4f} {:10.4f}".format(
            pave, pbias, prms, pstdev, pavestd
        )
    )

    return alpha_values, Pobs, Plow, Phigh, dPobs, Pnorm



##############################################################################
# pymbar: A Python Library for MBAR
#
# Copyright 2010-2017 University of Colorado Boulder, Memorial Sloan-Kettering Cancer Center
#
# Authors: Michael Shirts, John Chodera
# Contributors: Kyle Beauchamp, Levi Naden
#
# pymbar is free software: you can redistribute it and/or modify
# it under the terms of the MIT License.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# MIT License for more details.
#
# You should have received a copy of the MIT License along with pymbar.
##############################################################################

"""The pymbar package contains the pymbar suite of tools for the analysis of
simulated and experimental data with the multistate Bennett acceptance
ratio (MBAR) estimator.

"""

__author__ = "Michael R. Shirts and John D. Chodera"
__license__ = "MIT"
__maintainer__ = "Levi N. Naden, Jaime Rodríguez-Guerra, Michael R. Shirts and John D. Chodera"
__email__ = "levi.naden@choderalab.org,jaime.rodriguez-guerra@choderalab.org,michael.shirts@colorado.edu,john.chodera@choderalab.org"

from importlib.metadata import version, PackageNotFoundError

from . import timeseries, testsystems, confidenceintervals
from .mbar import MBAR
from .other_estimators import bar, bar_overlap, bar_zero, exp, exp_gauss
from .fes import FES


__all__ = [
    "exp",
    "exp_gauss",
    "bar",
    "bar_overlap",
    "bar_zero",
    "MBAR",
    "timeseries",
    "testsystems",
    "confidenceintervals",
    "utils",
    "FES",
]

try:
    __version__ = version("pymbar")
except PackageNotFoundError:
    # package is not installed
    pass



import numpy as np
import pytest
import pymbar
from pymbar.utils_for_testing import (
    assert_array_almost_equal,
    oscillators,
    exponentials,
)
from pymbar.tests.test_mbar import z_scale_factor
from pymbar.mbar_solvers import _setup_jax_acceleration


@pytest.fixture(scope="module")
def base_oscillator():
    name, u_kn, N_k, s_n, test = oscillators(50, 100, provide_test=True)
    return {"name": name, "u_kn": u_kn, "N_k": N_k, "s_n": s_n, "test": test}


@pytest.fixture(scope="module")
def more_oscillators():
    name, u_kn, N_k, s_n, test = oscillators(50, 500, provide_test=True)
    return {"name": name, "u_kn": u_kn, "N_k": N_k, "s_n": s_n, "test": test}


@pytest.mark.flaky(max_runs=2)  # Uses flaky plugin for pytest
@pytest.mark.parametrize(
    "statesa, statesb, test_system",
    [(100, 100, oscillators), (200, 50, oscillators), (200, 50, exponentials)],
)
def test_solvers(statesa, statesb, test_system):
    name, U, N_k, s_n, _ = test_system(statesa, statesb, provide_test=True)
    print(name)
    mbar = pymbar.MBAR(U, N_k)
    assert_array_almost_equal(
        pymbar.mbar_solvers.mbar_gradient(U, N_k, mbar.f_k), np.zeros(N_k.shape), decimal=8
    )
    assert_array_almost_equal(np.exp(mbar.Log_W_nk).sum(0), np.ones(len(N_k)), decimal=10)
    assert_array_almost_equal(np.exp(mbar.Log_W_nk).dot(N_k), np.ones(U.shape[1]), decimal=10)
    assert_array_almost_equal(
        pymbar.mbar_solvers.self_consistent_update(U, N_k, mbar.f_k), mbar.f_k, decimal=10
    )


def run_mbar_protocol(oscillator_bundle, protocol):
    test = oscillator_bundle["test"]
    u_kn = oscillator_bundle["u_kn"]
    N_k = oscillator_bundle["N_k"]
    fa = test.analytical_free_energies()
    fa = fa[1:] - fa[0]
    # Solve MBAR with zeros for initial weights
    mbar = pymbar.MBAR(u_kn, N_k, solver_protocol=({"method": protocol},))
    # Solve MBAR with the correct f_k used for the initial weights
    mbar = pymbar.MBAR(u_kn, N_k, initial_f_k=mbar.f_k, solver_protocol=({"method": protocol},))
    return mbar, fa


@pytest.mark.parametrize(
    "protocol",
    [
        "adaptive",
        "hybr",
        "lm",
        "L-BFGS-B",
        "dogleg",
        "CG",
        "BFGS",
        "Newton-CG",
        pytest.param("TNC", marks=pytest.mark.flaky(max_runs=4)),  # This one is flaky
        "trust-ncg",
        "trust-krylov",
        "trust-exact",
        "SLSQP",
    ],
)
def test_protocols(base_oscillator, more_oscillators, protocol):
    """
    Test that free energy is moderately equal to analytical solution, independent of solver protocols
    """

    # Importing the hacky fix to asert that free energies are moderately correct

    try:
        mbar, fa = run_mbar_protocol(base_oscillator, protocol)
    except Exception as e:  # pylint: disable=broad-except
        print(f"Caught error in initial oscillator test, trying with more samples. Error:\n\n{e}")
        mbar, fa = run_mbar_protocol(more_oscillators, protocol)
    results = mbar.compute_free_energy_differences()
    fe = results["Delta_f"][0, 1:]
    fe_sigma = results["dDelta_f"][0, 1:]
    z = (fe - fa) / fe_sigma
    assert_array_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


class TestPymbarJaxDisabled:
    @pytest.mark.parametrize(
        "env_value,expected",
        [
            ("true", True),
            ("TRUE", True),
            ("TrUe", True),
            ("yes", True),
            ("YES", True),
            ("YeS", True),
            ("1", True),
            ("false", False),
            ("FALSE", False),
            ("FaLsE", False),
            ("no", False),
            ("NO", False),
            ("nO", False),
            ("0", False),
            ("", False),
            ("invalid", False),
            ("random_value", False),
        ],
    )
    def test_env_values(self, monkeypatch, env_value, expected):
        """Test various environment variable values."""
        monkeypatch.setenv("PYMBAR_DISABLE_JAX", env_value)
        assert _setup_jax_acceleration() is expected

    def test_unset_variable(self, monkeypatch):
        """Test that unset variable returns False."""
        monkeypatch.delenv("PYMBAR_DISABLE_JAX", raising=False)
        assert _setup_jax_acceleration() is False



"""Test exp by performing statistical tests on a set of model systems
for which the true free energy differences can be computed analytically.
"""

import numpy as np
import pytest
from pymbar import other_estimators as estimators
from pymbar.testsystems import harmonic_oscillators, exponential_distributions
from pymbar.utils_for_testing import assert_equal, assert_almost_equal

precision = 8  # the precision for systems that do have analytical results that should be matched.
# Scales the z_scores so that we can reject things that differ at the ones decimal place.  TEMPORARY HACK
z_scale_factor = 12.0
# 0.5 is rounded to 1, so this says they must be within 3.0 sigma
N_k = np.array([50000, 100000])


def generate_ho(O_k=np.array([1.0, 2.0]), K_k=np.array([0.5, 1.0])):
    return "Harmonic Oscillators", harmonic_oscillators.HarmonicOscillatorsTestCase(O_k, K_k)


def generate_exp(rates=np.array([1.0, 4.0])):  # Rates, e.g. Lambda
    return "Exponentials", exponential_distributions.ExponentialTestCase(rates)


system_generators = [generate_ho, generate_exp]


@pytest.fixture(scope="module", params=system_generators)
def exp_and_test(request):
    name, test = request.param()
    w_F, w_R, N_k_output = test.sample(N_k, mode="wFwR")
    assert_equal(N_k, N_k_output)
    exps = dict()
    # can't return method, because exp is just a function
    exps["F"] = estimators.exp(w_F)
    exps["R"] = estimators.exp(w_R)
    exps["gF"] = estimators.exp_gauss(w_F)
    exps["gR"] = estimators.exp_gauss(w_R)

    yield_bundle = {"exps": exps, "test": test, "w_F": w_F, "w_R": w_R}
    yield yield_bundle


@pytest.mark.parametrize("system_generator", system_generators)
def test_sample(system_generator):
    """Draw samples via test object."""

    name, test = system_generator()
    print(name)

    w_F, w_R, N_k = test.sample([10, 8], mode="wFwR")
    w_F, w_R, N_k = test.sample([1, 1], mode="wFwR")
    w_F, w_R, N_k = test.sample([10, 0], mode="wFwR")
    w_F, w_R, N_k = test.sample([0, 5], mode="wFwR")


def test_EXP_free_energies(exp_and_test):
    """Can exp calculate moderately correct free energy differences?"""

    exps, test = exp_and_test["exps"], exp_and_test["test"]

    fe0 = test.analytical_free_energies()
    fe0 = fe0[1:] - fe0[0]

    results_F = exps["F"]
    fe_F = results_F["Delta_f"]
    dfe_F = results_F["dDelta_f"]
    z = (fe_F - fe0) / dfe_F
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)

    results_R = exps["R"]
    fe_R = -results_R["Delta_f"]
    dfe_R = results_R["dDelta_f"]
    z = (fe_R - fe0) / dfe_R
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)

    # turning off Gaussian comparisons because it's not clear how accurate they should be!

    results_gF = exps["gF"]
    fe_gF = results_gF["Delta_f"]
    dfe_gF = results_gF["dDelta_f"]
    z = (fe_gF - fe0) / dfe_gF
    # assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)

    results_gR = exps["gR"]
    fe_gR = -results_gR["Delta_f"]
    dfe_gR = results_gR["dDelta_f"]
    z = (fe_gR - fe0) / dfe_gR
    # assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)

    # make sure the different methods are nearly equal for these systems (within uncertainty)
    z = np.abs(fe_R - fe_F) / np.sqrt(dfe_R**2 + dfe_F**2)
    assert_almost_equal(z / z_scale_factor, 0.0, decimal=0)



__all__ = [
    "timeseries",
    "exponential_distributions",
    "harmonic_oscillators",
    "gaussian_work",
    "HarmonicOscillatorsTestCase",
    "ExponentialTestCase",
]

from pymbar.testsystems.harmonic_oscillators import HarmonicOscillatorsTestCase
from pymbar.testsystems.exponential_distributions import ExponentialTestCase
from pymbar.testsystems.timeseries import correlated_timeseries_example
from pymbar.testsystems.gaussian_work import gaussian_work_example



"""
Example illustrating the application of MBAR to compute a 1D free energy profile from an umbrella sampling simulation.

The data represents an umbrella sampling simulation for the chi torsion of
a valine sidechain in lysozyme L99A with benzene bound in the cavity.

Reference:

    D. L. Mobley, A. P. Graves, J. D. Chodera, A. C. McReynolds, B. K. Shoichet and K. A. Dill,
    "Predicting absolute ligand binding free energies to a simple model site,"
    Journal of Molecular Biology 371(4):1118-1134 (2007).
    http://dx.doi.org/10.1016/j.jmb.2007.06.002


"""

import numpy as np

import pymbar  # multistate Bennett acceptance ratio
from pymbar import timeseries  # timeseries analysis

# Constants.
kB = 1.381e-23 * 6.022e23 / 1000.0  # Boltzmann constant in kJ/mol/K

temperature = 300  # assume a single temperature -- can be overridden with data from center.dat

# Parameters
K = 26  # number of umbrellas
N_max = 501  # maximum number of snapshots/simulation
T_k = np.ones(K, float) * temperature  # inital temperatures are all equal
beta = 1.0 / (kB * temperature)  # inverse temperature of simulations (in 1/(kJ/mol))
chi_min = -180.0  # min for free energy profile
chi_max = +180.0  # max for free energy profile
nbins = 36  # number of bins for 1D free energy profile

# Allocate storage for simulation data
# N_k[k] is the number of snapshots from umbrella simulation k
N_k = np.zeros([K], dtype=int)
# K_k[k] is the spring constant (in kJ/mol/deg**2) for umbrella simulation k
K_k = np.zeros([K])
# chi0_k[k] is the spring center location (in deg) for umbrella simulation k
chi0_k = np.zeros([K])
# chi_kn[k,n] is the torsion angle (in deg) for snapshot n from umbrella simulation k
chi_kn = np.zeros([K, N_max])
# u_kn[k,n] is the reduced potential energy without umbrella restraints of snapshot n of umbrella simulation k
u_kn = np.zeros([K, N_max])
g_k = np.zeros([K])

# Read in umbrella spring constants and centers.
with open("data/centers.dat") as infile:
    lines = infile.readlines()

for k in range(K):
    # Parse line k.
    line = lines[k]
    tokens = line.split()
    chi0_k[k] = float(tokens[0])  # spring center location (in deg)
    # spring constant (read in kJ/mol/rad**2, converted to kJ/mol/deg**2)
    K_k[k] = float(tokens[1]) * (np.pi / 180) ** 2
    if len(tokens) > 2:
        T_k[k] = float(tokens[2])  # temperature the kth simulation was run at.

beta_k = 1.0 / (kB * T_k)  # beta factor for the different temperatures
different_temperatures = True
if min(T_k) == max(T_k):
    # if all the temperatures are the same, then we don't have to read in energies.
    different_temperatures = False

# Read the simulation data
for k in range(K):
    # Read torsion angle data.
    filename = f"data/prod{k:d}_dihed.xvg"
    print(f"Reading {filename}...")
    n = 0
    with open(filename, "r") as infile:
        for line in infile:
            if line[0] != "#" and line[0] != "@":
                tokens = line.split()
                chi = float(tokens[1])  # torsion angle
                # wrap chi_kn to be within [-180,+180)
                while chi < -180.0:
                    chi += 360.0
                while chi >= +180.0:
                    chi -= 360.0
                chi_kn[k, n] = chi
                n += 1
    N_k[k] = n

    if different_temperatures:  # if different temperatures are specified the metadata file,
        # then we need the energies to compute the free energy profile
        # Read energies
        filename = f"data/prod{k:d}_energies.xvg"
        print(f"Reading {filename}...")
        n = 0
        with open(filename, "r") as infile:
            for line in infile:
                if line[0] != "#" and line[0] != "@":
                    tokens = line.split()
                    # reduced potential energy without umbrella restraint
                    u_kn[k, n] = beta_k[k] * (float(tokens[2]) - float(tokens[1]))
                    n += 1

    # Compute correlation times for potential energy and chi
    # timeseries.  If the temperatures differ, use energies to determine samples; otherwise, use the cosine of chi

    if different_temperatures:
        g_k[k] = timeseries.statistical_inefficiency(u_kn[k, :], u_kn[k, 0 : N_k[k]])
        print(f"Correlation time for set {k:5d} is {g_k[k]:10.3f}")
        indices = timeseries.subsample_correlated_data(u_kn[k, 0 : N_k[k]])
    else:
        chi_radians = chi_kn[k, 0 : N_k[k]] / (180.0 / np.pi)
        g_cos = timeseries.statistical_inefficiency(np.cos(chi_radians))
        g_sin = timeseries.statistical_inefficiency(np.sin(chi_radians))
        print(f"g_cos = {g_cos:.1f} | g_sin = {g_sin:.1f}")
        g_k[k] = max(g_cos, g_sin)
        print(f"Correlation time for set {k:5d} is {g_k[k]:10.3f}")
        indices = timeseries.subsample_correlated_data(chi_radians, g=g_k[k])
    # Subsample data.
    N_k[k] = len(indices)
    u_kn[k, 0 : N_k[k]] = u_kn[k, indices]
    chi_kn[k, 0 : N_k[k]] = chi_kn[k, indices]

N_max = np.max(N_k)  # shorten the array size
# u_kln[k,l,n] is the reduced potential energy of snapshot n from umbrella simulation k evaluated at umbrella l
u_kln = np.zeros([K, K, N_max])

# Set zero of u_kn -- this is arbitrary.
u_kn -= u_kn.min()

# compute bin centers
bin_center_i = np.zeros([nbins])
bin_edges = np.linspace(chi_min, chi_max, nbins + 1)
for i in range(nbins):
    bin_center_i[i] = 0.5 * (bin_edges[i] + bin_edges[i + 1])

N = np.sum(N_k)
chi_n = pymbar.utils.kn_to_n(chi_kn, N_k=N_k)

# Evaluate reduced energies in all umbrellas
print("Evaluating reduced potential energies...")
for k in range(K):
    for n in range(N_k[k]):
        # Compute minimum-image torsion deviation from umbrella center l
        dchi = chi_kn[k, n] - chi0_k
        for l in range(K):
            if abs(dchi[l]) > 180.0:
                dchi[l] = 360.0 - abs(dchi[l])

        # Compute energy of snapshot n from simulation k in umbrella potential l
        u_kln[k, :, n] = u_kn[k, n] + beta_k[k] * (K_k / 2.0) * dchi**2

# initialize free energy profile with the data collected
fes = pymbar.FES(u_kln, N_k, verbose=True)
# Compute free energy profile in unbiased potential (in units of kT).
histogram_parameters = {}
histogram_parameters["bin_edges"] = bin_edges
fes.generate_fes(u_kn, chi_n, fes_type="histogram", histogram_parameters=histogram_parameters)
results = fes.get_fes(bin_center_i, reference_point="from-lowest", uncertainty_method="analytical")
center_f_i = results["f_i"]
center_df_i = results["df_i"]

# Write out free energy profile
print("free energy profile (in units of kT), from histogramming")
print(f"{'bin':>8s} {'f':>8s} {'df':>8s}")
for i in range(nbins):
    print(f"{bin_center_i[i]:8.1f} {center_f_i[i]:8.3f} {center_df_i[i]:8.3f}")

# NOW KDE:
kde_parameters = {}
kde_parameters["bandwidth"] = 0.5 * ((chi_max - chi_min) / nbins)
fes.generate_fes(u_kn, chi_n, fes_type="kde", kde_parameters=kde_parameters)
results = fes.get_fes(bin_center_i, reference_point="from-lowest")
# Write out free energy profile from KDE
center_f_i = results["f_i"]
print("")
print("free energy profile (in units of kT), from KDE")
print(f"{'bin':>8s} {'f':>8s}")
for i in range(nbins):
    print(f"{bin_center_i[i]:8.1f} {center_f_i[i]:8.3f}")



"""Test MBAR by performing statistical tests on a set of model systems
for which the true free energy differences can be computed analytically.
"""

import numpy as np
import pytest
from pymbar import MBAR
from pymbar.testsystems import harmonic_oscillators, exponential_distributions
from pymbar.utils_for_testing import assert_equal, assert_almost_equal
from pymbar.utils import ParameterError

precision = 8  # the precision for systems that do have analytical results that should be matched.
# Scales the z_scores so that we can reject things that differ at the ones decimal place.  TEMPORARY HACK
z_scale_factor = 12.0
# 0.5 is rounded to 1, so this says they must be within 3.0 sigma
N_k = np.array([1000, 500, 0, 800])


def generate_ho(O_k=np.array([1.0, 2.0, 3.0, 4.0]), K_k=np.array([0.5, 1.0, 1.5, 2.0])):
    return "Harmonic Oscillators", harmonic_oscillators.HarmonicOscillatorsTestCase(O_k, K_k)


def generate_exp(rates=np.array([1.0, 2.0, 3.0, 4.0])):  # Rates, e.g. Lambda
    return "Exponentials", exponential_distributions.ExponentialTestCase(rates)


def convert_to_differences(x_ij, dx_ij, xa):
    xa_ij = xa - np.vstack(xa)

    # add ones to the diagonal of the uncertainties, because they are zero
    for i in range(len(N_k)):
        dx_ij[i, i] += 1
    z = (x_ij - xa_ij) / dx_ij
    for i in range(len(N_k)):
        # these terms should be zero; so we only throw an error if they aren't
        z[i, i] = x_ij[i, i] - xa_ij[i, i]
    return z


system_generators = [generate_ho, generate_exp]
observables = ["position", "position^2", "RMS deviation", "potential energy"]


@pytest.fixture(scope="module", params=system_generators)
def mbar_and_test(request):
    name, test = request.param()
    x_n, u_kn, N_k_output, s_n = test.sample(N_k, mode="u_kn")
    assert_equal(N_k, N_k_output)
    mbar = MBAR(u_kn, N_k, verbose=True, n_bootstraps=200)  # Bootstrap needed for a few tests
    yield_bundle = {"mbar": mbar, "test": test, "x_n": x_n, "u_kn": u_kn}
    yield yield_bundle


@pytest.fixture(scope="module")
def mbar_and_test_harmonic():
    name, test = generate_ho()
    x_n, u_kn, N_k_output, s_n = test.sample(N_k, mode="u_kn")
    assert_equal(N_k, N_k_output)
    mbar = MBAR(u_kn, N_k, verbose=True)
    yield_bundle = {"mbar": mbar, "test": test, "x_n": x_n, "u_kn": u_kn}
    yield yield_bundle


@pytest.fixture(scope="module")
def mbar_and_test_kln():
    name, test = generate_ho()
    x_n, u_kn, N_k_output = test.sample(N_k, mode="u_kln")
    assert_equal(N_k, N_k_output)
    mbar = MBAR(u_kn, N_k, verbose=True)
    yield_bundle = {"mbar": mbar, "test": test, "x_n": x_n, "u_kn": u_kn}
    yield yield_bundle


@pytest.fixture()  # Function  level scope
def fixed_harmonic_sample():
    _, test = generate_ho()
    return test


@pytest.fixture()
def single_harmonic_u_kn(fixed_harmonic_sample):
    _, u_kn, _, _ = fixed_harmonic_sample.sample(N_k, mode="u_kn")
    return u_kn


def free_energies_almost_equal(mbar_fe, err_fe, analytical_fe):
    """Helper to test if MBAR vs Analytical free energies are almost equal"""
    mbar_fe, err_fe = mbar_fe[0, 1:], err_fe[0, 1:]
    analytical_fe = analytical_fe[1:] - analytical_fe[0]

    z = (mbar_fe - analytical_fe) / err_fe
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


def test_ukln(mbar_and_test_kln):
    """Test that MBAR's u_kln->u_kn works correctly"""
    mbar, test = mbar_and_test_kln["mbar"], mbar_and_test_kln["test"]
    results = mbar.compute_free_energy_differences()
    fe = results["Delta_f"]
    fe_sigma = results["dDelta_f"]
    free_energies_almost_equal(fe, fe_sigma, test.analytical_free_energies())


def test_duplicate_state(fixed_harmonic_sample, caplog):
    """Test that MBAR's duplicate state check is working"""
    _, u_kn, _, _ = fixed_harmonic_sample.sample(N_k, mode="u_kn")
    u_kn_dup = np.append(u_kn, u_kn[[0], :], axis=0)
    N_k_dup = np.append(N_k, [0])
    mbar = MBAR(u_kn_dup, N_k_dup, verbose=True)
    assert "likely to to be the same thermodynamic state" in caplog.text
    results = mbar.compute_free_energy_differences()
    fe = results["Delta_f"]
    assert np.allclose(fe[0], fe[-1])


def test_x_kindices(fixed_harmonic_sample):
    """Test that setting x_kindices results in the same calculation"""
    _, u_kn, _, _ = fixed_harmonic_sample.sample(N_k, mode="u_kn")
    flat_x_indices = np.concatenate([[k] * n for k, n in enumerate(N_k)]).astype(int)
    mbar = MBAR(u_kn, N_k)
    mbar_indices = MBAR(u_kn, N_k, x_kindices=flat_x_indices)
    fe = mbar.compute_free_energy_differences()["Delta_f"]
    fes = mbar_indices.compute_free_energy_differences()["Delta_f"]
    assert np.allclose(fe, fes)


@pytest.mark.xfail(raises=ParameterError)
def test_bad_inital_f_k(fixed_harmonic_sample):
    _, u_kn, _, _ = fixed_harmonic_sample.sample(N_k, mode="u_kn")
    MBAR(u_kn, N_k, initial_f_k=[0] * (N_k.size + 1))


def test_covariance_of_sums_runs(mbar_and_test_kln):
    """Test that CovarianceOfSums function runs"""
    # TODO: Is this function still needed? And what would be a better test?
    mbar = mbar_and_test_kln["mbar"]
    results = mbar.compute_free_energy_differences(return_theta=True)
    theta = results["Theta"]
    mbar.compute_covariance_of_sums(theta, 1, np.array([1, -1]))


@pytest.mark.parametrize("system_generator", system_generators)
def test_analytical(system_generator):
    """Generate test objects and calculate analytical results."""
    name, test = system_generator()
    mu = test.analytical_means()
    variance = test.analytical_variances()
    f_k = test.analytical_free_energies()
    for observable in observables:
        A_k = test.analytical_observable(observable=observable)
    s_k = test.analytical_entropies()


@pytest.mark.parametrize("system_generator", system_generators)
def test_sample(system_generator):
    """Draw samples via test object."""

    name, test = system_generator()
    print(name)

    x_n, u_kn, N_k, s_n = test.sample([5, 6, 7, 8], mode="u_kn")
    x_n, u_kn, N_k, s_n = test.sample([5, 5, 5, 5], mode="u_kn")
    x_n, u_kn, N_k, s_n = test.sample([1, 1, 1, 1], mode="u_kn")
    x_n, u_kn, N_k, s_n = test.sample([10, 0, 8, 0], mode="u_kn")

    x_kn, u_kln, N_k = test.sample([5, 6, 7, 8], mode="u_kln")
    x_kn, u_kln, N_k = test.sample([5, 5, 5, 5], mode="u_kln")
    x_kn, u_kln, N_k = test.sample([1, 1, 1, 1], mode="u_kln")
    x_kn, u_kln, N_k = test.sample([10, 0, 8, 0], mode="u_kln")


@pytest.mark.parametrize(
    "uncertainty_method",
    [
        None,
        "approximate",
        "svd",
        "svd-ew",
        "bootstrap",
        pytest.param("waffles", marks=pytest.mark.xfail),
    ],
)
def test_mbar_free_energies(mbar_and_test, uncertainty_method):
    """Can MBAR calculate moderately correct free energy differences?"""
    mbar, test = mbar_and_test["mbar"], mbar_and_test["test"]

    results = mbar.compute_free_energy_differences(
        return_theta=True, uncertainty_method=uncertainty_method
    )
    fe = results["Delta_f"]
    fe_sigma = results["dDelta_f"]
    free_energies_almost_equal(fe, fe_sigma, test.analytical_free_energies())


@pytest.mark.xfail(strict=True)  # This whole test should always fail and passes are problems
@pytest.mark.parametrize("n_bootstrap", [None, -4, 0, 100.3])
def test_mbar_bad_bootstrap(single_harmonic_u_kn, n_bootstrap):
    """Test that bad parameters for n_bootstraps makes bootstrap fail"""
    mbar = MBAR(single_harmonic_u_kn, N_k, verbose=True, n_bootstraps=n_bootstrap)
    mbar.compute_free_energy_differences(uncertainty_method="bootstrap")


@pytest.mark.parametrize(
    "method",
    ["zeros", "mean-reduced-potential", "BAR", pytest.param("waffles", marks=pytest.mark.xfail)],
)
def test_mbar_initialization(fixed_harmonic_sample, method):
    """Do the MBAR initialization methods work?"""
    _, u_kn, _, _ = fixed_harmonic_sample.sample(N_k, mode="u_kn")
    mbar = MBAR(u_kn, N_k, initialize=method, verbose=True)
    results = mbar.compute_free_energy_differences()
    fe = results["Delta_f"]
    fe_sigma = results["dDelta_f"]
    free_energies_almost_equal(fe, fe_sigma, fixed_harmonic_sample.analytical_free_energies())


def test_mbar_compute_expectations_position_averages(mbar_and_test):
    """Can MBAR calculate E(x_n)??"""

    mbar, test, x_n = mbar_and_test["mbar"], mbar_and_test["test"], mbar_and_test["x_n"]
    results = mbar.compute_expectations(x_n)
    mu = results["mu"]
    sigma = results["sigma"]

    mu0 = test.analytical_observable(observable="position")

    z = (mu0 - mu) / sigma
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


def test_mbar_compute_expectations_position_differences(mbar_and_test):
    """Can MBAR calculate E(x_n)??"""
    mbar, test, x_n = mbar_and_test["mbar"], mbar_and_test["test"], mbar_and_test["x_n"]
    results = mbar.compute_expectations(x_n, output="differences")
    mu_ij = results["mu"]
    sigma_ij = results["sigma"]

    mu0 = test.analytical_observable(observable="position")
    z = convert_to_differences(mu_ij, sigma_ij, mu0)
    assert_almost_equal(z / z_scale_factor, np.zeros(np.shape(z)), decimal=0)


def test_mbar_compute_expectations_position2(mbar_and_test):
    """Can MBAR calculate E(x_n^2)??"""

    mbar, test, x_n = mbar_and_test["mbar"], mbar_and_test["test"], mbar_and_test["x_n"]
    results = mbar.compute_expectations(x_n**2)
    mu = results["mu"]
    sigma = results["sigma"]
    mu0 = test.analytical_observable(observable="position^2")

    z = (mu0 - mu) / sigma
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


def test_mbar_compute_expectations_potential(mbar_and_test):
    """Can MBAR calculate E(u_kn)??"""

    mbar, test, u_kn = mbar_and_test["mbar"], mbar_and_test["test"], mbar_and_test["u_kn"]
    results = mbar.compute_expectations(u_kn, state_dependent=True)
    mu = results["mu"]
    sigma = results["sigma"]
    mu0 = test.analytical_observable(observable="potential energy")
    z = (mu0 - mu) / sigma
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


@pytest.mark.parametrize(
    "observable,state_dependent,sample_mode,single_dim,with_uxx",
    [
        ("position", False, "u_kln", False, False),
        ("position", False, "u_kln", False, True),
        ("position", False, "u_kn", False, False),
        ("position", False, "u_kn", False, True),
        ("position", False, "u_kn", True, False),
        ("potential energy", True, "u_kln", False, False),
        ("potential energy", True, "u_kln", False, True),
        ("potential energy", True, "u_kn", False, False),
        ("potential energy", True, "u_kn", False, True),
        ("potential energy", True, "u_kn", True, False),
    ],
)
def test_mbar_compute_expectations_edges(
    mbar_and_test_harmonic,
    mbar_and_test_kln,
    observable,
    state_dependent,
    sample_mode,
    single_dim,
    with_uxx,
):
    if sample_mode == "u_kln":
        payload = mbar_and_test_kln
    else:
        payload = mbar_and_test_harmonic
    mbar = payload["mbar"]
    test = payload["test"]
    u_xxx = payload["u_kn"]
    if state_dependent:
        obs = payload["u_kn"]
    else:
        obs = payload["x_n"]
    if single_dim:
        u_xxx = u_xxx[0]
    results = mbar.compute_expectations(
        obs, state_dependent=state_dependent, u_kn=u_xxx if with_uxx else None, return_theta=True
    )
    mu = results["mu"]
    sigma = results["sigma"]
    mu0 = test.analytical_observable(observable=observable)
    z = (mu0 - mu) / sigma
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


def multiExpectationAssertion(results, test, state=1):
    mu = results["mu"]
    sigma = results["sigma"]
    mu0 = test.analytical_observable(observable="position")[state]
    mu1 = test.analytical_observable(observable="position^2")[state]
    z = (mu0 - mu[0]) / sigma[0]
    assert_almost_equal(z / z_scale_factor, 0 * z, decimal=0)
    z = (mu1 - mu[1]) / sigma[1]
    assert_almost_equal(z / z_scale_factor, 0 * z, decimal=0)


def test_mbar_compute_multiple_expectations(mbar_and_test):
    """Can MBAR calculate E(u_kn)??"""

    mbar, test, x_n, u_kn = (
        mbar_and_test["mbar"],
        mbar_and_test["test"],
        mbar_and_test["x_n"],
        mbar_and_test["u_kn"],
    )
    A = np.zeros([2, len(x_n)])
    A[0, :] = x_n
    A[1, :] = x_n**2
    state = 1
    results = mbar.compute_multiple_expectations(A, u_kn[state, :])
    multiExpectationAssertion(results, test, state=state)


def test_mbar_compute_multiple_expectations_more_dims(mbar_and_test_kln):
    """Can MBAR calculate E(u_kn) with 3 dimensions??"""

    mbar, test, x_n, u_kn = (
        mbar_and_test_kln["mbar"],
        mbar_and_test_kln["test"],
        mbar_and_test_kln["x_n"],
        mbar_and_test_kln["u_kn"],
    )
    A = np.zeros([2, x_n.shape[0], x_n.shape[1]])
    A[0, :, :] = x_n
    A[1, :, :] = x_n**2
    state = 1
    results = mbar.compute_multiple_expectations(
        A, u_kn[:, state, :], compute_covariance=True, return_theta=True
    )
    multiExpectationAssertion(results, test, state=state)


def test_mbar_compute_entropy_and_enthalpy(mbar_and_test, with_uxx=True):
    """Can MBAR calculate f_k, <u_k> and s_k ??"""

    mbar, test, x_n, u_kn = (
        mbar_and_test["mbar"],
        mbar_and_test["test"],
        mbar_and_test["x_n"],
        mbar_and_test["u_kn"],
    )
    results = mbar.compute_entropy_and_enthalpy(u_kn if with_uxx else None, verbose=True)
    f_ij = results["Delta_f"]
    df_ij = results["dDelta_f"]
    u_ij = results["Delta_u"]
    du_ij = results["dDelta_u"]
    s_ij = results["Delta_s"]
    ds_ij = results["dDelta_s"]

    fa = test.analytical_free_energies()
    ua = test.analytical_observable("potential energy")
    sa = test.analytical_entropies()

    fa_ij = fa - fa.T
    ua_ij = ua - ua.T
    sa_ij = sa - sa.T

    z = convert_to_differences(f_ij, df_ij, fa)
    assert_almost_equal(z / z_scale_factor, np.zeros(np.shape(z)), decimal=0)
    z = convert_to_differences(u_ij, du_ij, ua)
    assert_almost_equal(z / z_scale_factor, np.zeros(np.shape(z)), decimal=0)
    z = convert_to_differences(s_ij, ds_ij, sa)
    assert_almost_equal(z / z_scale_factor, np.zeros(np.shape(z)), decimal=0)


@pytest.mark.parametrize(
    "as_kln,with_uxx",
    [
        (True, True),
        (True, False),
        (False, False),
        # False True handled by main test
    ],
)
def test_mbar_compute_entropy_and_enthalpy_edges(
    mbar_and_test_harmonic, mbar_and_test_kln, as_kln, with_uxx
):
    if as_kln:
        payload = mbar_and_test_kln
    else:
        payload = mbar_and_test_harmonic
    test_mbar_compute_entropy_and_enthalpy(payload, with_uxx=with_uxx)


def test_mbar_compute_effective_sample_number(mbar_and_test):
    """testing compute_effective_sample_number"""

    mbar = mbar_and_test["mbar"]
    # one mathematical effective sample numbers should be between N_k and sum_k N_k
    N_eff = mbar.compute_effective_sample_number()
    sumN = np.sum(N_k)
    assert all(N_eff > N_k)
    assert all(N_eff < sumN)


def test_mbar_compute_overlap_analytical():
    """Tests Overlap with identical states, which gives analytical results."""

    d = len(N_k)
    even_O_k = 2.0 * np.ones(d)
    even_K_k = 0.5 * np.ones(d)
    even_N_k = 100 * np.ones(d)
    name, test = generate_ho(O_k=even_O_k, K_k=even_K_k)
    x_n, u_kn, N_k_output, s_n = test.sample(even_N_k, mode="u_kn")
    mbar = MBAR(u_kn, even_N_k)

    results = mbar.compute_overlap()
    overlap_scalar = results["scalar"]
    eigenval = results["eigenvalues"]
    O = results["matrix"]

    reference_matrix = (1.0 / d) * np.ones([d, d])
    reference_eigenvalues = np.zeros(d)
    reference_eigenvalues[0] = 1.0
    reference_scalar = np.float64(1.0)

    assert_almost_equal(O, reference_matrix, decimal=precision)
    assert_almost_equal(eigenval, reference_eigenvalues, decimal=precision)
    assert_almost_equal(overlap_scalar, reference_scalar, decimal=precision)


def test_mbar_compute_overlap_nonanalytical(mbar_and_test):
    """Tests Overlap with stochastic tests"""
    mbar = mbar_and_test["mbar"]
    results = mbar.compute_overlap()
    overlap_scalar = results["scalar"]
    eigenval = results["eigenvalues"]
    O = results["matrix"]

    assert isinstance(overlap_scalar, (float, int))
    # rows of matrix should sum to one
    sumrows = np.array(np.sum(O, axis=1))
    assert_almost_equal(sumrows, np.ones(np.shape(sumrows)), decimal=precision)
    assert_almost_equal(eigenval[0], np.float64(1.0), decimal=precision)


def test_mbar_weights(mbar_and_test):
    """testing weights"""

    mbar = mbar_and_test["mbar"]
    W = mbar.weights()
    sumrows = np.sum(W, axis=0)
    assert_almost_equal(sumrows, np.ones(len(sumrows)), decimal=precision)


@pytest.mark.parametrize(
    "system_generator,mode,bad_n",
    [
        (generate_ho, "u_kn", False),
        (generate_exp, "u_kn", False),
        (generate_ho, "u_kln", False),
        pytest.param(generate_ho, "u_kn", True, marks=pytest.mark.xfail(strict=True)),
    ],
)
def test_mbar_computePerturbedFreeEnergeies(system_generator, mode, bad_n):
    """testing compute_perturbed_free_energies"""

    # only do MBAR with the first and last set

    name, test = system_generator()
    if mode == "u_kln":
        x_n, u_kn, N_k_output = test.sample(N_k, mode=mode)
        numN = max(N_k[:2])
        if bad_n:
            numN = numN - 1
        mslice = np.s_[:2, :2, :numN]
        pslice = np.s_[:2, 2:, :numN]
    else:
        x_n, u_kn, N_k_output, s_n = test.sample(N_k, mode=mode)
        numN = np.sum(N_k[:2])
        if bad_n:
            numN = numN - 1
        mslice = np.s_[:2, :numN]
        pslice = np.s_[2:, :numN]

    mbar = MBAR(u_kn[mslice], N_k[:2])
    results = mbar.compute_perturbed_free_energies(u_kn[pslice])
    fe = results["Delta_f"]
    fe_sigma = results["dDelta_f"]

    fe, fe_sigma = fe[0, 1:], fe_sigma[0, 1:]
    fe0 = test.analytical_free_energies()[2:]
    fe0 = fe0[1:] - fe0[0]

    z = (fe - fe0) / fe_sigma
    assert_almost_equal(z / z_scale_factor, np.zeros(len(z)), decimal=0)


def test_mbar_compute_expectations_inner(mbar_and_test):
    """Can MBAR calculate general expectations inner code (note: this just tests completion)"""

    mbar, test, x_n, u_kn = (
        mbar_and_test["mbar"],
        mbar_and_test["test"],
        mbar_and_test["x_n"],
        mbar_and_test["u_kn"],
    )
    A_in = np.array([x_n, x_n**2, x_n**3])
    u_n = u_kn[:2, :]
    state_map = np.array([[0, 0], [1, 0], [2, 0], [2, 1]], int)
    _ = mbar.compute_expectations_inner(A_in, u_n, state_map)


@pytest.mark.parametrize("n_bootstrap", [1, 100])
def test_mbar_bootstrap_deterministic_given_same_seed(fixed_harmonic_sample, n_bootstrap):
    """Verify that providing a seed to the mbar bootstrap will produce the same values"""
    _, u_kn, _, _ = fixed_harmonic_sample.sample(N_k, mode="u_kn")

    mbar_a = MBAR(u_kn, N_k, verbose=True, n_bootstraps=n_bootstrap, rseed=814)
    ref_fe_diff = mbar_a.compute_free_energy_differences(uncertainty_method="bootstrap")

    # The verbose flag should have no impact on the determinism
    mbar_b = MBAR(u_kn, N_k, verbose=False, n_bootstraps=n_bootstrap, rseed=814)
    test_fe_diff = mbar_b.compute_free_energy_differences(uncertainty_method="bootstrap")
    np.testing.assert_equal(ref_fe_diff["Delta_f"], test_fe_diff["Delta_f"])
    np.testing.assert_equal(ref_fe_diff["dDelta_f"], test_fe_diff["dDelta_f"])
